from panqec.decoders import (
    BaseDecoder, RotatedSweepDecoder3D, MatchingDecoder
)
from panqec.codes import StabilizerCode
from panqec.error_models import BaseErrorModel
import numpy as np


class RotatedSweepMatchDecoder(BaseDecoder):

    label = 'Rotated Planar Code 3D Sweep Matching Decoder'
    allowed_codes = ["RotatedToric3DCode", "RotatedPlanar3DCode"]

    def __init__(self, code: StabilizerCode,
                 error_model: BaseErrorModel,
                 error_rate: float,
                 max_rounds=32):
        super().__init__(code, error_model, error_rate)

        self.max_rounds = max_rounds

        self.sweeper = RotatedSweepDecoder3D(
            code, error_model, error_rate, max_rounds=max_rounds
        )
        self.matcher = MatchingDecoder(
            code, error_model, error_rate, 'X'
        )

    @property
    def params(self) -> dict:
        return {
            'max_rounds': self.max_rounds
        }

    def decode(self, syndrome: np.ndarray, **kwargs) -> np.ndarray:
        """Get X and Z corrections given code and measured syndrome."""

        z_correction = self.sweeper.decode(syndrome)
        x_correction = self.matcher.decode(syndrome)

        correction = (x_correction + z_correction) % 2

        return correction



import itertools
import pytest
import numpy as np
from panqec.codes import Toric3DCode
from panqec.decoders import SweepDecoder3D
from panqec.bpauli import bsf_wt
from panqec.error_models import PauliErrorModel
from panqec.utils import edge_coords, face_coords
from tests.decoders.decoder_test import DecoderTest


class TestSweepDecoder3D(DecoderTest):

    @pytest.fixture
    def code(self):
        return Toric3DCode(3, 4, 5)

    @pytest.fixture
    def error_model(self):
        return PauliErrorModel(0, 0, 1)

    @pytest.fixture
    def decoder(self, code, error_model):
        error_rate = 0.5
        return SweepDecoder3D(code, error_model, error_rate)

    @pytest.fixture
    def allowed_paulis(self):
        return ['Z']

    def test_decode_Z_error(self, decoder, code):
        error = code.to_bsf({
            (2, 1, 2): 'Z',
        })
        assert bsf_wt(error) == 1

        # Measure the syndrome and ensure non-triviality.
        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2
        assert np.all(code.measure_syndrome(total_error) == 0)

    def test_decode_many_Z_errors(self, decoder, code):
        error = dict()
        error = code.to_bsf({
            (1, 0, 0): 'Z',
            (0, 1, 0): 'Z',
            (0, 0, 3): 'Z',
        })
        assert bsf_wt(error) == 3

        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2
        assert np.all(code.measure_syndrome(total_error) == 0)

    def test_unable_to_decode_X_error(self, decoder, code):
        error = code.to_bsf({
            (1, 0, 2): 'X'
        })
        assert bsf_wt(error) == 1

        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        correction = decoder.decode(syndrome)
        assert np.all(correction == 0)

        total_error = (error + correction) % 2
        assert np.all(error == total_error)

        assert np.any(code.measure_syndrome(total_error) != 0)

    def test_decode_many_codes_and_errors_with_same_decoder(self):

        codes = [
            Toric3DCode(3, 4, 5),
            Toric3DCode(3, 3, 3),
            Toric3DCode(5, 4, 3),
        ]

        sites = [
            (0, 0, 1),
            (1, 0, 0),
            (0, 1, 0)
        ]

        error_model = PauliErrorModel(0, 0, 1)
        error_rate = 0.5

        for code, site in itertools.product(codes, sites):
            decoder = SweepDecoder3D(code, error_model, error_rate)
            error = code.to_bsf({
                site: 'Z',
            })
            syndrome = code.measure_syndrome(error)
            correction = decoder.decode(syndrome)
            total_error = (error + correction) % 2
            assert np.all(code.measure_syndrome(total_error) == 0)


class Test3x3x3SweepDecoder3D(DecoderTest):

    @pytest.fixture
    def code(self):
        return Toric3DCode(3, 3, 3)

    @pytest.fixture
    def decoder(self, code, error_model):
        error_rate = 0.1
        return SweepDecoder3D(code, error_model, error_rate)

    @pytest.fixture
    def error_model(self):
        return PauliErrorModel(0, 0, 1)

    @pytest.fixture
    def allowed_paulis(self):
        return ['Z']

    def test_decode_error_on_two_edges_sharing_same_vertex(
        self, code, decoder
    ):
        error = code.to_bsf({
            (1, 0, 0): 'Z',
            (0, 1, 0): 'Z',
        })
        syndrome = code.measure_syndrome(error)
        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2
        assert np.all(code.measure_syndrome(total_error) == 0)

    def test_decode_with_general_Z_noise(self):
        code = Toric3DCode(3, 3, 3)
        error_model = PauliErrorModel(0, 0, 1)
        error_rate = 0.1
        decoder = SweepDecoder3D(code, error_model, error_rate)
        np.random.seed(0)

        in_codespace = []
        for i in range(100):
            error = error_model.generate(
                code, error_rate=error_rate, rng=np.random
            )
            syndrome = code.measure_syndrome(error)
            correction = decoder.decode(syndrome)
            total_error = (error + correction) % 2
            in_codespace.append(
                np.all(code.measure_syndrome(total_error) == 0)
            )

        # Some will just be fails and not in code space, but assert that at
        # least some of them ended up in the code space.
        assert any(in_codespace)

    @pytest.mark.parametrize(
        'edge_location, faces_flipped',
        [
            ((1, 0, 0), {
                (1, 1, 0), (1, 5, 0), (1, 0, 1), (1, 0, 5)
            }),
            ((0, 1, 0), {
                (1, 1, 0), (5, 1, 0), (0, 1, 1), (0, 1, 5)
            }),
            ((0, 0, 1), {
                (1, 0, 1), (5, 0, 1), (0, 1, 1), (0, 5, 1)
            }),
        ]
    )
    def test_flip_edge(self, edge_location, faces_flipped, code, decoder):
        signs = decoder.get_initial_state(
            np.zeros(code.stabilizer_matrix.shape[0])
        )
        decoder.flip_edge(edge_location, signs)
        assert faces_flipped == {
            code.stabilizer_coordinates[index]
            for index in np.where(signs)[0]
        }

    def test_decode_loop_step_by_step(self, code, decoder):
        sites = set(edge_coords([
            (0, 0, 0, 0), (1, 1, 0, 0), (0, 0, 1, 0), (1, 0, 0, 0),
        ], code.size))

        error_pauli = dict()
        for site in sites:
            error_pauli[site] = 'Z'
        assert set(error_pauli.keys()) == set(sites)
        error = code.to_bsf(error_pauli)

        # Initialize the correction.
        correction = dict()

        # Compute the syndrome.
        syndrome = code.measure_syndrome(error)

        signs = decoder.get_initial_state(syndrome)
        assert np.all(signs == syndrome)
        # assert np.all(
        #     rebuild_syndrome(code, signs)[:code.n_k_d[0]]
        #     == syndrome[:code.n_k_d[0]]
        # )

        # assert np.all(signs.reshape(code.n) == syndrome[:code.n])
        assert {
            code.stabilizer_coordinates[index]
            for index in np.where(signs)[0]
        } == set(face_coords([
            (0, 0, 0, 0), (0, 0, 0, 2), (0, 1, 0, 0), (0, 1, 0, 2),
            (1, 0, 0, 0), (1, 0, 0, 2), (1, 0, 1, 0), (1, 0, 1, 2),
            (2, 0, 1, 0), (2, 0, 2, 0), (2, 1, 0, 0), (2, 2, 0, 0),
        ], code.size))

        signs = decoder.sweep_move(signs, correction)
        assert set(correction.keys()) == set(edge_coords([
            (0, 0, 1, 0), (1, 1, 0, 0),
            (2, 0, 0, 0), (2, 0, 0, 2),
        ], code.size))
        assert {
            code.stabilizer_coordinates[index]
            for index in np.where(signs)[0]
        } == set(face_coords([
            (0, 0, 2, 0), (0, 0, 2, 2),
            (1, 2, 0, 0), (1, 2, 0, 2),
            (2, 2, 0, 0), (2, 0, 2, 0),
        ], code.size))

        signs = decoder.sweep_move(signs, correction)
        assert set(correction.keys()) == set(edge_coords([
            (0, 0, 1, 0), (1, 1, 0, 0),
            (2, 0, 0, 0), (2, 0, 0, 2),
            (0, 2, 0, 0), (1, 0, 2, 0)
        ], code.size))
        assert np.all(signs == 0)

        total_error = (error + code.to_bsf(correction)) % 2
        vertex_operator = code.get_stabilizer((0, 0, 0))
        assert np.all(total_error == code.to_bsf(vertex_operator))

        assert np.all(code.measure_syndrome(total_error) == 0)

    def test_decode_loop_ok(self, code, decoder):

        error_pauli = dict()
        """
        sites = [
            (0, 0, 0, 0), (1, 1, 0, 0), (0, 0, 1, 0), (1, 0, 0, 0)
        ]
        """
        sites = [(1, 0, 0), (2, 1, 0), (1, 2, 0), (0, 1, 0)]
        for site in sites:
            error_pauli[site] = 'Z'
        assert set(error_pauli.keys()) == set(sites)
        error = code.to_bsf(error_pauli)

        # Compute the syndrome.
        syndrome = code.measure_syndrome(error)

        signs = decoder.get_initial_state(syndrome)
        """
        assert dict_where(signs) == {
            (0, 0, 0, 0), (0, 0, 0, 2), (0, 1, 0, 0), (0, 1, 0, 2),
            (1, 0, 0, 0), (1, 0, 0, 2), (1, 0, 1, 0), (1, 0, 1, 2),
            (2, 0, 1, 0), (2, 0, 2, 0), (2, 1, 0, 0), (2, 2, 0, 0)
        }
        """
        assert {
            code.stabilizer_coordinates[index]
            for index in np.where(signs)[0]
        } == {
            (1, 2, 1), (2, 1, 1), (1, 0, 1), (3, 1, 0), (1, 5, 0), (1, 0, 5),
            (0, 1, 5), (1, 2, 5), (2, 1, 5), (5, 1, 0), (1, 3, 0), (0, 1, 1)
        }

        assert np.all(signs == syndrome)

        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2

        assert np.all(code.measure_syndrome(total_error) == 0)

    def test_oscillating_cycle_fail(self, code, decoder):

        sites = edge_coords([
            (0, 0, 1, 0), (0, 0, 1, 1), (0, 0, 2, 0), (0, 0, 2, 1),
            (0, 0, 2, 2), (0, 1, 1, 2), (0, 2, 0, 0), (0, 2, 0, 1),
            (0, 2, 0, 2), (1, 1, 0, 1), (1, 1, 2, 0), (1, 1, 2, 2),
            (1, 2, 0, 0), (1, 2, 0, 1), (1, 2, 0, 2), (1, 2, 2, 0),
            (1, 2, 2, 1), (1, 2, 2, 2), (2, 1, 0, 0), (2, 1, 1, 1),
            (2, 1, 2, 1),
        ], code.size)
        assert len(set(sites)) == 21

        error_pauli = dict()
        for site in sites:
            error_pauli[site] = 'Z'
        error = code.to_bsf(error_pauli)

        syndrome = code.measure_syndrome(error)

        # Signs array.
        signs = decoder.get_initial_state(syndrome)

        # Keep a copy of the initial signs array.
        start_signs = signs.copy()

        # Keep track of the correction to apply.
        correction = dict()

        # Sweep 3 times.
        for i_sweep in range(3):
            signs = decoder.sweep_move(signs, correction)

        # Back to the start again.
        assert np.all(signs == start_signs)

        # The total correction is trivial.
        assert np.all(
            code.measure_syndrome(code.to_bsf(correction)) == 0
        )

        # The total error still is not in code space.
        total_error = (error + code.to_bsf(correction)) % 2
        assert np.any(code.measure_syndrome(total_error) != 0)

    def test_never_ending_staircase_fails(self, code, decoder):

        # Weight-8 Z error that may start infinite loop in sweep decoder.
        error_pauli = dict()
        sites = edge_coords([
            (0, 0, 2, 2), (0, 1, 1, 1), (0, 2, 0, 2), (1, 0, 0, 0),
            (1, 1, 0, 2), (1, 2, 2, 1), (2, 1, 2, 1), (2, 2, 0, 0)
        ], code.size)
        for site in sites:
            error_pauli[site] = 'Z'
        error = code.to_bsf(error_pauli)
        # assert error.sum() == 8

        # Compute the syndrome and make sure it's nontrivial.
        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome)

        # Check face X stabilizer syndrome measurements.
        expected_syndrome_faces = face_coords([
            (0, 0, 0, 0), (0, 0, 0, 2), (0, 1, 0, 1), (0, 1, 0, 2),
            (0, 1, 1, 1), (0, 1, 2, 1), (0, 2, 0, 0), (0, 2, 2, 1),
            (1, 0, 2, 2), (1, 1, 0, 0), (1, 1, 1, 0), (1, 1, 1, 1),
            (1, 1, 2, 1), (1, 2, 0, 0), (1, 2, 0, 1), (1, 2, 0, 2),
            (2, 0, 0, 0), (2, 0, 0, 2), (2, 0, 1, 2), (2, 0, 2, 2),
            (2, 1, 0, 1), (2, 1, 0, 2), (2, 1, 1, 1), (2, 1, 2, 1),
            (2, 2, 0, 0), (2, 2, 0, 2), (2, 2, 2, 1), (2, 2, 2, 2)
        ], code.size)
        expected_signs = {k: 0 for k in code.type_index('face')}
        for k in expected_syndrome_faces:
            expected_signs[k] = 1
        expected_syndrome = rebuild_syndrome(
            code, expected_signs
        )
        assert np.all(syndrome == expected_syndrome)
        """
        assert np.all(
            np.array(expected_syndrome_faces).T
            == np.where(syndrome[:code.n].reshape(3, 3, 3, 3))
        )
        """

        # Attempt to perform decoding.
        correction = decoder.decode(syndrome)

        total_error = (error + correction) % 2

        # Assert that decoding has failed.
        np.any(code.measure_syndrome(total_error))

    def test_sweep_move_two_edges(self, code, decoder):

        error = {
            (0, 1, 0): 'Z',
            (1, 0, 0): 'Z',
        }

        syndrome = code.measure_syndrome(code.to_bsf(error))

        correction = dict()

        # Syndrome from errors on x edge and y edge on vertex (0, 0, 0).
        signs = {k: 0 for k in code.type_index('face')}
        faces = [
            (1, 5, 0), (0, 1, 1), (0, 1, 5),
            (5, 1, 0), (1, 0, 1), (1, 0, 5),
        ]
        signs = np.zeros(code.stabilizer_matrix.shape[0], dtype=np.uint)
        for face in faces:
            signs[code.stabilizer_index[face]] = 1

        assert np.all(decoder.get_initial_state(syndrome) == signs)

        # Expected signs after one sweep.
        expected_faces_1 = [
            (0, 5, 1), (0, 5, 5),
            (1, 5, 0),
            (5, 0, 1), (5, 0, 5),
            (5, 1, 0),
        ]
        signs_1 = decoder.sweep_move(signs, correction)
        faces_1 = {
            code.stabilizer_coordinates[index]
            for index in np.where(signs_1)[0]
        }
        assert set(expected_faces_1) == set(faces_1)

        # Expected signs after two sweeps, should be all gone.
        signs_2 = decoder.sweep_move(signs_1, correction)
        assert all(signs_2 == 0)

        expected_correction = {
            (0, 5, 0): 'Z',
            (5, 0, 0): 'Z',
            (0, 0, 1): 'Z',
            (0, 0, 5): 'Z',
        }

        assert correction == expected_correction


def find_sites(error_pauli):
    """List of sites where Pauli has support over."""
    return set([
        location
        for location, index in error_pauli.code.qubit_index.items()
        if index in np.where(error_pauli._zs.toarray()[0])[0]
    ])


def rebuild_syndrome(code, signs):
    reconstructed_syndrome = np.zeros(
        code.stabilizer_matrix.shape[0], dtype=np.uint
    )
    for location, index in code.type_index('face').items():
        if signs[location]:
            reconstructed_syndrome[index] = 1
    return reconstructed_syndrome



from typing import Tuple
import numpy as np
import pytest
from panqec.codes import RotatedPlanar3DCode
from panqec.decoders import RotatedSweepMatchDecoder
from panqec.error_models import PauliErrorModel
from tests.decoders.decoder_test import DecoderTest


class TestDeformedRotatedPlanarMatchingDecoder(DecoderTest):

    direction: Tuple[float, float, float] = 0.05, 0.05, 0.9
    size: Tuple[int, int, int] = 4, 4, 3
    error_rate: float = 0.1

    @pytest.fixture
    def code(self):
        return RotatedPlanar3DCode(*self.size)

    @pytest.fixture
    def error_model(self):
        return PauliErrorModel(
            *self.direction,
            deformation_name='XZZX',
            deformation_kwargs={'deformation_axis': 'z'}
        )

    @pytest.fixture
    def decoder(self, code, error_model):
        return RotatedSweepMatchDecoder(code, error_model,
                                        self.error_rate)

    def test_decode(self, code, error_model, decoder):
        rng = np.random.default_rng(seed=0)
        error = error_model.generate(code, self.error_rate, rng=rng)
        assert np.any(error != 0)

        syndrome = code.measure_syndrome(error)
        correction = decoder.decode(syndrome)
        total_error = (correction + error) % 2
        effective_error = code.logical_errors(total_error)
        codespace = bool(
            np.all(code.measure_syndrome(total_error) == 0)
        )
        assert codespace, 'Not in code space'
        success = bool(np.all(effective_error == 0)) and codespace
        assert success, 'Decoding failed'

    def test_decode_along_line_preferred(self, code, error_model, decoder):

        # Two close-together parallel lines of X errors along deformation axis.
        error = code.to_bsf({
            (4, 2, 2): 'X',
            (4, 2, 4): 'X',
            (6, 4, 2): 'X',
            (6, 4, 4): 'X',
        })

        # The expected correction, given the high bias, should be matching
        # along the deformation axis, which is the original error itself.
        expected_correction = error

        # (as opposed to the naive correction of joining the two ends across to
        # form a total error that is a loop)
        pauli_naive_correction = dict()
        pauli_naive_correction[(5, 3, 1)] = 'X'
        pauli_naive_correction[(5, 3, 5)] = 'X'
        naive_correction = code.to_bsf(pauli_naive_correction)

        syndrome = code.measure_syndrome(error)
        correction = decoder.decode(syndrome)

        assert np.any(correction != naive_correction), 'Correction is naive'
        assert np.all(correction == expected_correction), (
            'Correction not as expected'
        )
        total_error = (correction + error) % 2

        effective_error = code.logical_errors(total_error)
        codespace = code.in_codespace(total_error)
        assert codespace, 'Not in code space'
        success = bool(np.all(effective_error == 0)) and codespace
        assert success, 'Decoding failed'



from ._toric_2d_code import Toric2DCode  # noqa
from ._planar_2d_code import Planar2DCode  # noqa
from ._rotated_planar_2d_code import RotatedPlanar2DCode  # noqa

__all__ = [
    'Toric2DCode',
    'Planar2DCode',
    'RotatedPlanar2DCode',
]



from typing import Tuple, Dict
import numpy as np
from panqec.decoders import BaseDecoder
from panqec.codes import StabilizerCode
from panqec.error_models import BaseErrorModel

Operator = Dict[Tuple, str]


class RotatedSweepDecoder3D(BaseDecoder):

    label = 'Rotated Code 3D Sweep Decoder'
    allowed_codes = ["RotatedToric3DCode", "RotatedPlanar3DCode"]

    _rng: np.random.Generator
    max_rounds: int

    def __init__(self, code: StabilizerCode,
                 error_model: BaseErrorModel,
                 error_rate: float,
                 seed: int = 0,
                 max_rounds: int = 32):
        super().__init__(code, error_model, error_rate)
        self._rng = np.random.default_rng(seed)
        self.seed = seed
        self.max_rounds = max_rounds

    @property
    def params(self) -> dict:
        return {
            'seed': self.seed,
            'max_rounds': self.max_rounds
        }

    def get_face_syndromes(
        self, full_syndrome: np.ndarray
    ) -> np.ndarray:
        """Get only the syndromes for the vertex Z stabilizers.
        Z vertex stabilizers syndromes are discarded for this decoder.
        """
        face_syndromes = self.code.extract_x_syndrome(full_syndrome)
        return face_syndromes

    # TODO: make this more space-efficient, don't store zeros.
    def get_initial_state(
        self, syndrome: np.ndarray
    ) -> np.ndarray:
        """Get initial cellular automaton state from syndrome."""
        signs = syndrome.copy()
        signs[self.code.z_indices] = 0

        return signs

    def decode(
        self, syndrome: np.ndarray, **kwargs
    ) -> np.ndarray:
        """Get Z corrections given measured syndrome."""

        # Maximum number of times to apply sweep rule before giving up round.
        largest_size = 2*int(max(self.code.size)) + 2
        max_sweeps = 4*largest_size

        # The syndromes represented as an array of 0s and 1s.
        signs = self.get_initial_state(syndrome)

        # Keep track of the correction needed.
        correction: Dict = dict()

        # Sweep directions to take
        sweep_directions = [
            (1, 0, 1), (1, 0, -1),
            (0, 1, 1), (0, 1, -1),
            (-1, 0, 1), (-1, 0, -1),
            (0, -1, 1), (0, -1, -1),
        ]

        # Keep sweeping in all directions until there are no syndromes.
        i_round = 0
        while any(signs) and i_round < self.max_rounds:
            for sweep_direction in sweep_directions:

                # Initialize the number of sweeps.
                i_sweep = 0

                # Keep sweeping until there are no syndromes.
                while any(signs) and i_sweep < max_sweeps:
                    signs = self.sweep_move(
                        signs, correction, sweep_direction
                    )
                    i_sweep += 1
            i_round += 1

        return self.code.to_bsf(correction)

    def get_sweep_faces(self, vertex, sweep_direction):
        """Get the coordinates of neighboring faces in sweep direction."""
        x, y, z = vertex
        s_x, s_y, s_z = sweep_direction

        # (s_x, s_y) in [(1, 0), (0, 1)]
        if s_x + s_y > 0:
            x_face = (x + 1, y + 1, z + 1*s_z)

        # (s_x, s_y) in [(-1, 0), (0, -1)]
        else:
            x_face = (x - 1, y - 1, z + 1*s_z)

        # (s_x, s_y) in [(1, 0), (0, -1)]
        if s_x - s_y > 0:
            y_face = (x + 1, y - 1, z + 1*s_z)

        # (s_x, s_y) in [(-1, 0), (0, 1)]
        else:
            y_face = (x - 1, y + 1, z + 1*s_z)

        z_face = (x + 2*s_x, y + 2*s_y, z)
        return x_face, y_face, z_face

    def get_sweep_edges(self, vertex, sweep_direction):
        """Get coordinates of neighbouring edges in sweep direction."""
        x, y, z = vertex
        s_x, s_y, s_z = sweep_direction

        if s_x - s_y > 0:
            x_edge = (x + 1, y - 1, z)
        else:
            x_edge = (x - 1, y + 1, z)

        if s_x + s_y > 0:
            y_edge = (x + 1, y + 1, z)
        else:
            y_edge = (x - 1, y - 1, z)

        z_edge = (x, y, z + 1*s_z)

        return x_edge, y_edge, z_edge

    def get_default_direction(self):
        """The default direction when all faces are excited."""
        direction = int(self._rng.choice([0, 1, 2], size=1))
        return direction

    def sweep_move(
        self, signs: np.ndarray, correction: Operator,
        sweep_direction: Tuple[int, int, int]
    ) -> np.ndarray:
        """Apply the sweep move once along a particular direciton."""

        flip_locations = []

        # Apply sweep rule on every vertex.
        for vertex in self.code.stabilizer_coordinates:
            if self.code.stabilizer_type(vertex) == 'vertex':
                # Get neighbouring faces and edges in the sweep direction.
                x_face, y_face, z_face = self.get_sweep_faces(
                    vertex, sweep_direction
                )
                x_edge, y_edge, z_edge = self.get_sweep_edges(
                    vertex, sweep_direction
                )

                # Check faces and edges are in lattice before proceeding.
                faces_valid = tuple(
                    self.code.is_stabilizer(face, 'face')
                    for face in [x_face, y_face, z_face]
                )
                edges_valid = tuple(
                    edge in self.code.qubit_index
                    for edge in [x_edge, y_edge, z_edge]
                )
                if all(faces_valid) and all(edges_valid):

                    x_face_index = self.code.stabilizer_index[x_face]
                    y_face_index = self.code.stabilizer_index[y_face]
                    z_face_index = self.code.stabilizer_index[z_face]
                    if (
                        signs[x_face_index] and signs[y_face_index]
                        and signs[z_face_index]
                    ):
                        direction = self.get_default_direction()
                        edge_flip = {
                            0: x_edge, 1: y_edge, 2: z_edge
                        }[direction]
                        flip_locations.append(edge_flip)
                    elif signs[y_face_index] and signs[z_face_index]:
                        flip_locations.append(x_edge)
                    elif signs[x_face_index] and signs[z_face_index]:
                        flip_locations.append(y_edge)
                    elif signs[x_face_index] and signs[y_face_index]:
                        flip_locations.append(z_edge)

                """
                # Boundary case with only 1 face and 2 edges.
                elif sum(faces_valid) == 1 and sum(edges_valid) == 2:
                    i_face_valid = faces_valid.index(True)
                    i_edge_invalid = edges_valid.index(False)
                    i_edge_valid = edges_valid.index(True)
                    # Check valid face is orthogonal to missing edge.
                    if i_face_valid == i_edge_invalid:
                        valid_face = [x_face, y_face, z_face][i_face_valid]
                        valid_edge = [x_edge, y_edge, z_edge][i_edge_valid]
                        # Flip an edge if the face has a syndrome.
                        if signs[valid_face]:
                            flip_locations.append(valid_edge)
                """

        new_signs = signs.copy()
        for location in flip_locations:
            self.flip_edge(location, new_signs)
            self.code.site(correction, 'Z', location)

        return new_signs

    def flip_edge(self, edge: Tuple, signs: np.ndarray):
        """Flip signs at index and update correction."""

        x, y, z = edge

        # Determine the axis the edge is parallel to.
        if z % 2 == 0:
            edge_direction = 'z'
        elif x % 4 == 1:
            if y % 4 == 1:
                edge_direction = 'x'
            elif y % 4 == 3:
                edge_direction = 'y'
        elif x % 4 == 3:
            if y % 4 == 1:
                edge_direction = 'y'
            elif y % 4 == 3:
                edge_direction = 'x'

        # Get the faces adjacent to the edge.
        if edge_direction == 'x':
            faces = [
                (x + 1, y + 1, z),
                (x - 1, y - 1, z),
                (x, y, z + 1),
                (x, y, z - 1),
            ]
        elif edge_direction == 'y':
            faces = [
                (x + 1, y - 1, z),
                (x - 1, y + 1, z),
                (x, y, z + 1),
                (x, y, z - 1),
            ]
        elif edge_direction == 'z':
            faces = [
                (x + 1, y + 1, z),
                (x - 1, y - 1, z),
                (x - 1, y + 1, z),
                (x + 1, y - 1, z),
            ]

        # Only keep faces that are actually on the cut lattice.
        faces = [face for face in faces
                 if self.code.is_stabilizer(face, 'face')]

        # Flip the state of the faces.
        for face in faces:
            signs[self.code.stabilizer_index[face]] = 1 - signs[
                self.code.stabilizer_index[face]
            ]



from typing import Tuple
from abc import ABCMeta, abstractmethod
import numpy as np
from panqec.codes import StabilizerCode


class BaseErrorModel(metaclass=ABCMeta):
    """Base class for error models"""

    @property
    @abstractmethod
    def label(self) -> str:
        """Label used in plots result files
        E.g. 'PauliErrorModel X1 Y0 Z0'
        """

    @property
    def id(self) -> str:
        return self.__class__.__name__

    @property
    @abstractmethod
    def params(self) -> dict:
        """List of class arguments (as a dictionary), that can be saved
        and reused to instantiate the same error model"""

    @abstractmethod
    def generate(
        self, code: StabilizerCode, error_rate: float, rng=None
    ) -> np.ndarray:
        """Generate errors for a given code and probability of failure

        Parameters
        ----------
        code : StabilizerCode
            Errors will be generated on the qubits of the provided code
        error_rate: float
            Physical error rate
        rng: numpy.random.Generator
            Random number generator (default=None resolves to
            numpy.random.default_rng())

        Returns
        -------
        error : np.ndarray
            Error as an array of size 2n (with n the number of qubits)
            in the binary symplectic format
        """

    @abstractmethod
    def probability_distribution(
        self, code: StabilizerCode, error_rate: float
    ) -> Tuple:
        """Probability distribution of X, Y and Z errors on all the
        qubits of a code. Can be used to generate errors and configure
        decoders.

        Parameters
        ----------
        code : StabilizerCode
            Code used for the error model
        error_rate: float
            Physical error rate

        Returns
        -------
        p_i, p_x, p_y, p_z : Tuple[np.ndarray]
            Probability distribution for I, X, Y and Z errors.
            Each probability is an array of size n (number of qubits)
        """

    def error_probability(
        self, error: np.ndarray, code: StabilizerCode,
        error_rate: float, log_output: bool = False
    ) -> np.ndarray:

        pi, px, py, pz = self.probability_distribution(code, error_rate)

        prob_vector = np.zeros(code.n)
        prob_vector += py * (error[:code.n] == error[code.n:])
        prob_vector += px * np.logical_and(error[:code.n],
                                           np.logical_not(error[code.n:]))
        prob_vector += pz * np.logical_and(np.logical_not(error[:code.n]),
                                           error[code.n:])
        prob_vector += pi * np.logical_and(np.logical_not(error[:code.n]),
                                           np.logical_not(error[code.n:]))

        if log_output:
            prob = np.sum(np.log(prob_vector))
        else:
            prob = np.prod(prob_vector)

        return prob

    def get_weights(
        self, code: StabilizerCode, error_rate: float, eps=1e-20
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Get MWPM weights for deformed Pauli noise."""

        pi, px, py, pz = self.probability_distribution(
            code, error_rate
        )

        total_p_x = px + py
        total_p_z = pz + py

        weights_x = -np.log(
            (total_p_x + eps) / (1 - total_p_x + eps)
        )
        weights_z = -np.log(
            (total_p_z + eps) / (1 - total_p_z + eps)
        )

        return weights_x, weights_z



from ._base_simulation import BaseSimulation  # noqa
from ._direct_simulation import (  # noqa
    DirectSimulation, calculate_logical_error_rate, run_once
)
from ._splitting_simulation import SplittingSimulation  # noqa
from ._batch_simulation import (  # noqa
    BatchSimulation, read_input_json,
    read_input_dict, run_file,
    expand_input_ranges, count_runs,
)

__all__ = [
    'BaseSimulation',
    'DirectSimulation', 'BatchSimulation', 'SplittingSimulation',
    'run_file', 'read_input_json', 'read_input_dict', 'run_once',
]



from typing import Tuple, Dict, List
import numpy as np
from panqec.codes import StabilizerCode

Operator = Dict[Tuple, str]  # Location to pauli ('X', 'Y' or 'Z')
Coordinates = List[Tuple]  # List of locations


class RotatedPlanar3DCode(StabilizerCode):
    """3D surface code with open boundaries on lattice rotated about z axis.

    Uses roughly half as many qubits as
    :class:`panqec.codes.surface_3d.Planar3DCode`.

    Parameters
    ----------
    L_x : int
        Number of qubits in the x direction.
    L_y : Optional[int
        Number of qubits in the y direction.
    L_z : Optional[int]
        Number of qubits in the z direction.

    Notes
    -----
    The lattice is stacked with lattices like those in
    :class:`panqec.codes.surface_2d.RotatedPlanar2DCode`
    glued with vertical qubits in between each layer.
    """
    dimension = 3
    deformation_names = ['XZZX']

    @property
    def label(self) -> str:
        return 'Rotated Planar {}x{}x{}'.format(*self.size)

    def get_qubit_coordinates(self) -> Coordinates:
        Lx, Ly, Lz = self.size

        coordinates: Coordinates = []

        # Horizontal
        for x in range(1, 2*Lx, 2):
            for y in range(1, 2*Ly, 2):
                for z in range(1, 2*Lz, 2):
                    coordinates.append((x, y, z))

        # Vertical
        for x in range(2, 2*Lx, 2):
            for y in range(0, 2*Ly + 1, 2):
                for z in range(2, 2*Lz, 2):
                    if (x + y) % 4 == 2:
                        coordinates.append((x, y, z))
        return coordinates

    def get_stabilizer_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly, Lz = self.size

        # Vertices
        for x in range(2, 2*Lx, 2):
            for y in range(0, 2*Ly+1, 2):
                for z in range(1, 2*Lz, 2):
                    if (x + y) % 4 == 2:
                        coordinates.append((x, y, z))

        # Horizontal faces
        for x in range(0, 2*Lx+1, 2):
            for y in range(2, 2*Ly, 2):
                for z in range(1, 2*Lz, 2):
                    if (x + y) % 4 == 0:
                        coordinates.append((x, y, z))

        # Vertical faces
        for x in range(1, 2*Lx+1, 2):
            for y in range(1, 2*Ly, 2):
                for z in range(2, 2*Lz, 2):
                    coordinates.append((x, y, z))
        return coordinates

    def stabilizer_type(self, location: Tuple) -> str:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location} for a stabilizer")

        x, y, z = location
        if (x + y) % 4 == 2 and z % 2 == 1:
            return 'vertex'
        else:
            return 'face'

    def get_stabilizer(self, location) -> Operator:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location} for a stabilizer")

        if self.stabilizer_type(location) == 'vertex':
            pauli = 'Z'
        else:
            pauli = 'X'

        x, y, z = location

        if self.stabilizer_type(location) == 'vertex':
            delta = [
                (-1, -1, 0), (-1, 1, 0), (1, -1, 0), (1, 1, 0), (0, 0, -1),
                (0, 0, 1)
            ]
        else:
            # z-normal so face is xy-plane.
            if z % 2 == 1:
                delta = [(-1, -1, 0), (1, 1, 0), (-1, 1, 0), (1, -1, 0)]
            # x-normal so face is in yz-plane.
            elif (x + y) % 4 == 0:
                delta = [(-1, -1, 0), (1, 1, 0), (0, 0, -1), (0, 0, 1)]
            # y-normal so face is in zx-plane.
            elif (x + y) % 4 == 2:
                delta = [(-1, 1, 0), (1, -1, 0), (0, 0, -1), (0, 0, 1)]

        operator = dict()
        for d in delta:
            qubit_location = tuple(np.add(location, d))

            if self.is_qubit(qubit_location):
                operator[qubit_location] = pauli

        return operator

    def qubit_axis(self, location):
        x, y, z = location

        if location not in self.qubit_coordinates:
            raise ValueError(
                f'Location {location} does not correspond to a qubit'
            )

        if (z % 2 == 0):
            axis = 'z'
        elif (x + y) % 4 == 2:
            axis = 'x'
        elif (x + y) % 4 == 0:
            axis = 'y'

        return axis

    def get_logicals_x(self) -> List[Operator]:
        """Get the unique logical X operator."""
        Lx, Ly, Lz = self.size
        logicals = []

        # X operators along x edges in x direction.
        operator: Operator = dict()
        for x in range(1, 2*Lx, 2):
            operator[(x, 1, 1)] = 'X'
        logicals.append(operator)

        return logicals

    def get_logicals_z(self) -> List[Operator]:
        """Get the unique logical Z operator."""

        Lx, Ly, Lz = self.size
        logicals = []

        # Z operators along x edges in x direction.
        operator: Operator = dict()

        for z in range(1, 2*Lz, 2):
            for y in range(1, 2*Ly, 2):
                operator[(1, y, z)] = 'Z'
        logicals.append(operator)

        return logicals

    def qubit_representation(
        self, location, rotated_picture=False, json_file=None
    ) -> Dict:
        representation = super().qubit_representation(
            location, rotated_picture, json_file
        )

        if self.qubit_axis(location) == 'z':
            representation['params']['length'] = 2

        if rotated_picture:
            x, y, z = representation['location']
            representation['location'] = (x, y, z*1.4142)

        return representation

    def stabilizer_representation(
        self, location, rotated_picture=False, json_file=None
    ) -> Dict:
        representation = super().stabilizer_representation(
            location, rotated_picture, json_file
        )

        x, y, z = location
        if not rotated_picture and self.stabilizer_type(location) == 'face':
            if z % 2 == 1:
                representation['params']['normal'] = [0, 0, 1]
                representation['params']['angle'] = np.pi/4
            else:
                representation['params']['w'] = 1.5
                representation['params']['angle'] = 0

                if (x + y) % 4 == 0:
                    representation['params']['normal'] = [1, 1, 0]
                else:
                    representation['params']['normal'] = [-1, 1, 0]

        if rotated_picture and self.stabilizer_type(location) == 'face':
            if z % 2 == 1:
                representation['params']['normal'] = [0, 0, 1]
                representation['params']['angle'] = 0
            else:
                representation['params']['angle'] = np.pi/4

                if (x + y) % 4 == 0:
                    representation['params']['normal'] = [1, 1, 0]
                else:
                    representation['params']['normal'] = [-1, 1, 0]

        if rotated_picture:
            x, y, z = representation['location']
            representation['location'] = (x, y, z*1.4142)

        return representation

    def get_deformation(
        self, location: Tuple,
        deformation_name: str,
        deformation_axis: str = 'z',
        **kwargs
    ) -> Dict:

        if deformation_axis not in ['x', 'y', 'z']:
            raise ValueError(f"{deformation_axis} is not a valid "
                             "deformation axis")

        undeformed_dict = {'X': 'X', 'Y': 'Y', 'Z': 'Z'}

        if deformation_name == 'XZZX':
            deformed_dict = {'X': 'Z', 'Y': 'Y', 'Z': 'X'}

        else:
            raise ValueError(f"The deformation {deformation_name}"
                             "does not exist")

        if self.qubit_axis(location) == deformation_axis:
            return deformed_dict
        else:
            return undeformed_dict



import numpy as np
from typing import List, Dict
from panqec.decoders import BaseDecoder, MatchingDecoder
from panqec.codes import StabilizerCode, Toric2DCode
from panqec.decoders import BeliefPropagationOSDDecoder
from panqec.error_models import PauliErrorModel


def get_matched_pairs(H, correction, syndrome):
    pairs = []
    seen_syndromes = set([])
    syndrome_indices = np.nonzero(syndrome)[0]

    for s in syndrome_indices:
        if s not in seen_syndromes:
            # print("s", s)
            seen_syndromes.add(s)

            s_prime = s
            continue_search = True
            prev_qubit = -1

            while continue_search:
                found_new_qubit = False
                for q in H[s_prime].nonzero()[1]:
                    # print("q", q)
                    if correction[q] and q != prev_qubit:
                        # print("Found new qubit")
                        found_new_qubit = True
                        prev_qubit = q
                        for i in H[:, q].nonzero()[0]:
                            if i != s_prime:
                                s_prime = i
                                break

                        break

                if not found_new_qubit:
                    continue_search = False
                    pairs.append((s, s_prime))
                    seen_syndromes.add(s_prime)

    return pairs


def find_connected_components(neighbors):
    connected_components = []
    seen = set()

    def component(node):
        nodes_in_component = set([node])
        nodes = set([node])
        while nodes:
            node = nodes.pop()
            seen.add(node)
            nodes.update(neighbors[node] - seen)
            nodes_in_component.add(node)

        return list(nodes_in_component)

    for node in neighbors:
        if node not in seen:
            connected_components.append(component(node))

    return connected_components


def modular_average(numbers, weights, n):
    if isinstance(numbers, list):
        numbers = np.array(numbers)
    if isinstance(weights, list):
        weights = np.array(weights)

    weights = weights.reshape(-1, 1)

    circle_points = np.vstack([np.cos(2 * np.pi * numbers / n),
                               np.sin(2 * np.pi * numbers / n)]).T

    mass_center = np.mean(weights * circle_points, axis=0)
    if np.all(np.isclose(mass_center, 0)):
        avg = np.dot(numbers, weights)
    else:
        circle_avg = mass_center / np.linalg.norm(mass_center)
        avg = np.arccos(circle_avg[0]) * n / (2 * np.pi)

        if circle_avg[1] < 0:
            avg *= -1

    if avg < 0:
        avg += n

    return avg


def decode_plane(loops, size):
    Lx, Ly = size

    # Distinguish the two sides of the loops
    state = {}
    for x in range(0, 2*Lx, 2):
        current_state = 0
        for y in range(0, 2*Ly, 2):
            if y == 0:
                if (x - 1, y) in loops:
                    current_state = 1 - state[(x - 2, y)]
                elif x >= 2:
                    current_state = state[(x - 2, y)]
            if (x, y - 1) in loops:
                current_state = 1 - current_state

            # print((x, y), current_state)
            state[(x, y)] = current_state

    # Correct the minority vote
    count = np.unique(list(state.values()), return_counts=True)

    if len(count[0]) == 1:
        minority_state = 1 - count[0]
    else:
        minority_state = count[0][np.argmin(count[1])]

    # print(minority_state)

    correction_coordinates = []
    for x in range(0, 2*Lx, 2):
        for y in range(0, 2*Ly, 2):
            if state[(x, y)] == minority_state:
                correction_coordinates.append((x, y))

    return correction_coordinates


def get_toric_loop(xcube_matching_ortho, component, proj_axis_int):
    toric_loop_dict = {}
    for loc in xcube_matching_ortho:
        if loc[proj_axis_int] in component:
            loc_2d = tuple_remove(loc, proj_axis_int)

            if loc_2d in toric_loop_dict.keys():
                toric_loop_dict[loc_2d] += 1
            else:
                toric_loop_dict[loc_2d] = 1

    toric_loop = [coord for coord in toric_loop_dict.keys()
                  if toric_loop_dict[coord] % 2 == 1]

    return toric_loop


def tuple_insert(t, index, element):
    new_t = list(t)
    new_t.insert(index, element)
    return tuple(new_t)


def tuple_remove(t, index):
    new_t = list(t)
    new_t.pop(index)
    return tuple(new_t)


if __name__ == "__main__":
    avg = modular_average([1, 3, 5, 7], [0.5, 0.0, 0.5, 0.0], 8)

    print(avg)


class XCubeMatchingDecoder(BaseDecoder):
    """Matching decoder for 2D Toric Code, based on PyMatching"""

    label = 'XCube Matching'
    allowed_codes = ["XCubeCode"]

    def __init__(self,
                 code: StabilizerCode,
                 error_model: PauliErrorModel,
                 error_rate: float):
        """Constructor for the MatchingDecoder class

        Parameters
        ----------
        code : StabilizerCode
            Code used by the decoder
        error_model: PauliErrorModel
            Error model used by the decoder (to find the weights)
        error_rate: int, optional
            Error rate used by the decoder (to find the weights)
        error_type: str, optional
            Determines which type of errors (X or Z) to decode.
            Can take the values "X", "Z", or None if we want to
            decode all errors
        """
        super().__init__(code, error_model, error_rate)

        Lx, Ly, Lz = code.size

        self.toric_code = {'x': Toric2DCode(Ly, Lz),
                           'y': Toric2DCode(Lx, Lz),
                           'z': Toric2DCode(Lx, Ly)
                           }

        # Weight the 2D toric code matching decoders
        # Only works for Z biased noise and z-axis deformation
        weights_X, _ = self.error_model.get_weights(self.code,
                                                    self.error_rate)

        wz = weights_X[self.code.qubit_index[(0, 0, 1)]]
        wxy = weights_X[self.code.qubit_index[(1, 0, 0)]]

        weights = {
            'x': np.array([wxy if self.toric_code['x'].qubit_axis(loc) == 'x'
                           else wz
                           for loc in self.toric_code['x'].qubit_coordinates]
                          ),
            'y': np.array([wxy if self.toric_code['y'].qubit_axis(loc) == 'x'
                           else wz
                           for loc in self.toric_code['y'].qubit_coordinates]
                          ),
            'z': np.array([wxy
                           for _ in self.toric_code['z'].qubit_coordinates]
                          )}

        self.matching_decoder = {axis: MatchingDecoder(self.toric_code[axis],
                                                       self.error_model,
                                                       self.error_rate,
                                                       weights=(weights[axis],
                                                                weights[axis]))
                                 for axis in ['x', 'y', 'z']}

        self.z_decoder = BeliefPropagationOSDDecoder(self.code,
                                                     self.error_model,
                                                     self.error_rate)

    @property
    def params(self) -> dict:
        return {}

    def decode(self, syndrome: np.ndarray, **kwargs) -> np.ndarray:
        """Get X corrections given code and measured syndrome."""

        # Initialize correction as full bsf.
        possible_correction = {'x': np.zeros(2*self.code.n, dtype=np.uint),
                               'y': np.zeros(2*self.code.n, dtype=np.uint),
                               'z': np.zeros(2*self.code.n, dtype=np.uint)}

        Lx, Ly, Lz = self.code.size

        # print("Initialize plane syndrome")
        plane_syndrome = {
            'x': {x: np.zeros(self.toric_code['x'].n_stabilizers)
                  for x in range(1, 2*Lx, 2)},
            'y': {y: np.zeros(self.toric_code['y'].n_stabilizers)
                  for y in range(1, 2*Ly, 2)},
            'z': {z: np.zeros(self.toric_code['z'].n_stabilizers)
                  for z in range(1, 2*Lz, 2)}
        }

        # Remove X stabilizer syndrome and keep it for later
        x_syndrome = self.code.extract_x_syndrome(syndrome)
        syndrome[self.code.x_indices] = 0
        axis_to_int = {'x': 0, 'y': 1, 'z': 2}

        for proj_axis in ['x', 'y', 'z']:
            proj_axis_int = axis_to_int[proj_axis]

            ortho_axes = tuple_remove(('x', 'y', 'z'), proj_axis_int)

            L_proj = [Lx, Ly, Lz][proj_axis_int]

            for i_stab in range(len(self.code.stabilizer_index)):
                if syndrome[i_stab]:
                    loc = self.code.stabilizer_coordinates[i_stab]

                    for axis in ['x', 'y', 'z']:
                        loc_2d = tuple_remove(loc, axis_to_int[axis])
                        idx_face = self.toric_code[axis].stabilizer_index[
                            loc_2d
                        ]
                        plane = loc[axis_to_int[axis]]
                        plane_syndrome[axis][plane][idx_face] = 1

            # Decode all the 2D toric codes
            xcube_matching: Dict[str, List] = {'x': [], 'y': [], 'z': []}
            connected_planes: Dict[int, set] = {
                plane: set() for plane in plane_syndrome[proj_axis].keys()
            }

            for axis in ['x', 'y', 'z']:
                for plane in plane_syndrome[axis].keys():
                    current_syndrome = plane_syndrome[axis][plane]

                    toric_X_syndrome = self.toric_code[
                        axis
                    ].extract_x_syndrome(
                        current_syndrome
                    )

                    if np.all(toric_X_syndrome == 0):
                        continue

                    toric_matching = self.matching_decoder[axis].decode(
                        current_syndrome
                    )
                    n = self.toric_code[axis].n
                    toric_Z_correction = toric_matching[n:]

                    toric_pairs = get_matched_pairs(
                        self.toric_code[axis].Hx.todense(),
                        toric_Z_correction,
                        toric_X_syndrome
                    )

                    for i in toric_Z_correction.nonzero()[0]:
                        x, y = self.toric_code[axis].qubit_coordinates[i]
                        loc_3d = tuple_insert((x, y), axis_to_int[axis], plane)
                        xcube_matching[axis].append(loc_3d)

                    if axis != proj_axis:
                        for pair in toric_pairs:
                            i1, i2 = pair
                            loc1 = self.toric_code[axis].qubit_coordinates[i1]
                            loc2 = self.toric_code[axis].qubit_coordinates[i2]

                            # print("Axis", axis)
                            # print("Pair", loc1, loc2)

                            proj_component = {'x': {'y': 0, 'z': 0},
                                              'y': {'x': 0, 'z': 1},
                                              'z': {'x': 1, 'y': 1}
                                              }[proj_axis][axis]
                            plane1 = loc1[proj_component]
                            plane2 = loc2[proj_component]

                            if plane1 != plane2:
                                if proj_component == 1:
                                    connected_planes[plane1+1].add(plane2+1)
                                    connected_planes[plane2+1].add(plane1+1)
                                else:
                                    connected_planes[plane1].add(plane2)
                                    connected_planes[plane2].add(plane1)

            xcube_matching_proj = xcube_matching[proj_axis]
            xcube_matching_ortho = set((xcube_matching[ortho_axes[0]]
                                        + xcube_matching[ortho_axes[1]]))

            connected_components = find_connected_components(connected_planes)

            list_plane_proj = []
            for component in list(connected_components):
                plane_proj = component[0]
                list_plane_proj.append(plane_proj)

                # Perform the projection
                for loc in xcube_matching_proj:
                    # print("Z matching", (x, y, z))
                    plane = loc[proj_axis_int]
                    ortho_components = tuple_remove(loc, proj_axis_int)
                    if plane in component:
                        if (0 < plane - plane_proj <= L_proj or
                                2*L_proj + plane - plane_proj <= L_proj):
                            if plane_proj < plane:
                                plane_stop = plane + 1
                            else:
                                plane_stop = 2*L_proj + plane + 1

                            for plane_op in range(plane_proj+1, plane_stop, 2):
                                loc = tuple_insert(ortho_components,
                                                   proj_axis_int,
                                                   plane_op % (2*L_proj))

                                idx = self.code.qubit_index[loc]
                                possible_correction[proj_axis][idx] += 1
                        elif plane != plane_proj:
                            if plane < plane_proj:
                                plane_stop = plane_proj + 1
                            else:
                                plane_stop = 2*L_proj + plane_proj + 1

                            for plane_op in range(plane + 1, plane_stop, 2):
                                loc = tuple_insert(ortho_components,
                                                   proj_axis_int,
                                                   plane_op % (2*L_proj))

                                idx = self.code.qubit_index[loc]
                                possible_correction[proj_axis][idx] += 1

            # Find the loops
            for component in connected_components:
                toric_loop = get_toric_loop(
                    xcube_matching_ortho, component, proj_axis_int
                )
                correction_coordinates = decode_plane(toric_loop, (Lx, Ly))

                plane_proj = component[0]

                for loc_2d in correction_coordinates:
                    loc_3d = tuple_insert(loc_2d, proj_axis_int, plane_proj)
                    idx = self.code.qubit_index[loc_3d]

                    possible_correction[proj_axis][idx] = 1

        weight = [np.sum(c) for c in possible_correction.values()]
        index_min_weight = int(np.argmin(weight))

        axis_min_weight = {0: 'x', 1: 'y', 2: 'z'}[index_min_weight]
        correction = possible_correction[axis_min_weight]

        # Decode Z part with BP-OSD
        syndrome[self.code.z_indices] = 0
        syndrome[self.code.x_indices] = x_syndrome

        z_correction = self.z_decoder.decode(syndrome).astype('uint8')

        correction += z_correction

        return correction % 2



import itertools
import pytest
import numpy as np
from panqec.bpauli import bsf_wt
from panqec.codes import Toric3DCode
from panqec.decoders import SweepMatchDecoder
from panqec.error_models import PauliErrorModel
from tests.decoders.decoder_test import DecoderTest


class TestSweepMatchDecoder(DecoderTest):

    @pytest.fixture
    def code(self):
        return Toric3DCode(3, 4, 5)

    @pytest.fixture
    def decoder(self, code, error_model):
        error_rate = 0.5
        return SweepMatchDecoder(code, error_model, error_rate)

    def test_decoder_has_required_attributes(self, decoder):
        assert decoder.label is not None
        assert decoder.decode is not None

    def test_decode_trivial_syndrome(self, decoder, code):
        syndrome = np.zeros(
            shape=code.stabilizer_matrix.shape[0], dtype=np.uint
        )
        correction = decoder.decode(syndrome)
        assert correction.shape == (2*code.n,)
        assert np.all(code.measure_syndrome(correction) == 0)
        assert issubclass(correction.dtype.type, np.integer)

    @pytest.mark.parametrize(
        'pauli, location',
        [
            ('X', (5, 4, 4)),
            ('Y', (4, 1, 6)),
            ('Z', (2, 4, 7)),
        ]
    )
    def test_decode_single_error(self, decoder, code, pauli, location):
        error = code.to_bsf({
            location: pauli
        })
        assert bsf_wt(error) == 1

        # Measure the syndrome and ensure non-triviality.
        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2
        assert np.all(code.measure_syndrome(total_error) == 0)

    @pytest.mark.parametrize(
        'paulis_locations',
        [
            [
                ('X', (5, 4, 4)),
                ('Y', (4, 1, 6)),
                ('Z', (2, 4, 7)),
            ],
            [
                ('Y', (5, 2, 2)),
                ('Y', (4, 1, 6)),
                ('Y', (2, 4, 7)),
            ],
        ]
    )
    def test_decode_many_errors(self, decoder, code, paulis_locations):
        error = code.to_bsf({
            location: pauli
            for pauli, location in paulis_locations
        })
        assert bsf_wt(error) == len(paulis_locations)

        syndrome = code.measure_syndrome(error)
        assert np.any(syndrome != 0)

        correction = decoder.decode(syndrome)
        total_error = (error + correction) % 2
        assert np.all(code.measure_syndrome(total_error) == 0)

    def test_decode_many_codes_and_errors_with_same_decoder(self):

        codes = [
            Toric3DCode(3, 4, 5),
            Toric3DCode(3, 3, 3),
            Toric3DCode(5, 4, 3),
        ]

        sites = [
            (4, 5, 4),
            (3, 0, 4),
            (2, 2, 1)
        ]

        error_model = PauliErrorModel(1/3, 1/3, 1/3)
        error_rate = 0.5

        for code, site in itertools.product(codes, sites):
            decoder = SweepMatchDecoder(code, error_model, error_rate)
            error = code.to_bsf({
                site: 'Z'
            })
            syndrome = code.measure_syndrome(error)
            correction = decoder.decode(syndrome)
            total_error = (error + correction) % 2
            assert np.all(code.measure_syndrome(total_error) == 0)



import numpy as np
from typing import Optional, Tuple
from pymatching import Matching
from panqec.decoders import BaseDecoder
from panqec.codes import StabilizerCode
from panqec.error_models import BaseErrorModel


class MatchingDecoder(BaseDecoder):
    """Matching decoder for 2D Toric Code, based on PyMatching"""

    label = 'Toric 2D Matching'
    allowed_codes = ["Toric2DCode", "Planar2DCode", "RotatedPlanar2DCode"]

    def __init__(self,
                 code: StabilizerCode,
                 error_model: BaseErrorModel,
                 error_rate: float,
                 error_type: Optional[str] = None,
                 weights: Optional[Tuple[np.ndarray, np.ndarray]] = None):
        """Constructor for the MatchingDecoder class

        Parameters
        ----------
        code : StabilizerCode
            Code used by the decoder
        error_model: BaseErrorModel
            Error model used by the decoder (to find the weights)
        error_rate: int, optional
            Error rate used by the decoder (to find the weights)
        error_type: str, optional
            Determines which type of errors (X or Z) to decode.
            Can take the values "X", "Z", or None if we want to
            decode all errors
        weights: Tuple[np.ndarray, np.ndarray], optional
            Matching weights for X and Z errors
        """
        super().__init__(code, error_model, error_rate)

        if error_type not in ["X", "Z", None]:
            raise ValueError("Argument 'error_type' has to be 'X', 'Z'"
                             f"or None, not {error_type}")

        self.error_type = error_type
        self.weights = weights

        if weights is not None:
            wx, wz = weights
        else:
            wx, wz = error_model.get_weights(code, error_rate)

        if error_type is None or error_type == "X":
            self.matcher_x = Matching(self.code.Hz, spacelike_weights=wx)
        if error_type is None or error_type == "Z":
            self.matcher_z = Matching(self.code.Hx, spacelike_weights=wz)

    @property
    def params(self) -> dict:
        return {
            'error_type': self.error_type,
            'weights': self.weights
        }

    def decode(self, syndrome: np.ndarray, **kwargs) -> np.ndarray:
        """Get X corrections given code and measured syndrome."""

        # Initialize correction as full bsf.
        correction = np.zeros(2*self.code.n, dtype=np.uint)

        # Keep only the vertex Z measurement syndrome, discard the rest.
        if self.error_type is None or self.error_type == "X":
            syndromes_z = self.code.extract_z_syndrome(syndrome)
            correction_x = self.matcher_x.decode(syndromes_z,
                                                 num_neighbours=None)
            correction[:self.code.n] = correction_x
        if self.error_type is None or self.error_type == "Z":
            syndromes_x = self.code.extract_x_syndrome(syndrome)
            correction_z = self.matcher_z.decode(syndromes_x,
                                                 num_neighbours=None)
            correction[self.code.n:] = correction_z

        return correction



import pytest
from panqec.codes import RhombicPlanarCode
from tests.codes.stabilizer_code_test import StabilizerCodeTest


class TestRhombicPlanarCode(StabilizerCodeTest):

    @pytest.fixture
    def code(self):
        return RhombicPlanarCode(4)

    def test_stabilizer_index(self, code):
        assert all(
            len(index) in [3, 4] for index in code.stabilizer_index
        )



"""
Utilities for automating slurm jobs.

:Author:
    Eric Huang
"""

import os
import subprocess
import re
import datetime
from typing import Optional
from glob import glob
from .config import (
    SLURM_DIR, SLURM_USERNAME
)
from .simulation import read_input_json, count_runs


def _delete_files_with_ext(folder_path, extension):
    """Delete files with extension in folder."""
    files = glob(os.path.join(folder_path, f'*.{extension}'))
    total_files = len(files)
    for file_path in files:
        print(f'Deleting {file_path}')
        os.remove(file_path)
    print(f'Deleted {total_files} .{extension} files')


def clear_sbatch_folder():
    """Delete all sbatch files in the sbatch folder."""
    sbatch_dir = os.path.join(SLURM_DIR, 'sbatch')
    _delete_files_with_ext(sbatch_dir, 'sbatch')
    _delete_files_with_ext(sbatch_dir, 'sh')


def clear_out_folder():
    """Delete all out files in the sbatch folder."""
    sbatch_dir = os.path.join(SLURM_DIR, 'out')
    _delete_files_with_ext(sbatch_dir, 'out')


def count_input_runs(name: str) -> Optional[int]:
    input_dir = os.path.join(SLURM_DIR, 'inputs')
    input_file = os.path.join(input_dir, f'{name}.json')
    n_runs = count_runs(input_file)
    return n_runs


def write_submit_sh(name, sbatch_files, flags=''):
    """Write script to submit all sbatch files for a given run."""
    sbatch_dir = os.path.join(SLURM_DIR, 'sbatch')
    sh_path = os.path.join(sbatch_dir, f'submit_{name}.sh')
    n_jobs = len(sbatch_files)
    lines = [
        '#!/bin/bash',
        f'echo "Submitting {n_jobs} jobs for {name}"'
    ]
    for sbatch_file in sbatch_files:
        lines.append(
            f'sbatch {flags} \'{sbatch_file}\''
        )
    lines.append('echo "Success! Happy waiting!"')
    sh_contents = '\n'.join(lines)
    with open(sh_path, 'w') as f:
        f.write(sh_contents)
    print(f'Run {sh_path} to submit all jobs')


def get_status():
    """Get status of running jobs on slurm."""
    get_squeue_status()
    get_out_status()
    get_results_status()


def get_squeue_status():
    try:
        status_command = 'squeue'
        if SLURM_USERNAME is not None:
            status_command = f'squeue -u {SLURM_USERNAME}'
        squeue = subprocess.Popen(
            status_command.split(),
            stdout=subprocess.PIPE
        )
        squeue_output, squeue_error = squeue.communicate()
        print(squeue_output.decode('utf-8'))
    except Exception:
        print('No squeue status available.')


def get_out_status():
    out_dir = os.path.join(SLURM_DIR, 'out')
    out_dir_glob = os.path.join(out_dir, '*.out')
    out_files = glob(out_dir_glob)
    if len(out_files) > 0:
        for out_file in out_files:
            tail_command = f'tail -n2 {out_file}'
            out_status = subprocess.Popen(
                tail_command.split(),
            )
            try:
                while True:
                    line = out_status.stdout.readline()
                    if not line:
                        break
            except AttributeError:
                print('')
    else:
        print('No slurm .out files available')


def get_results_status():
    sbatch_files = glob(os.path.join(SLURM_DIR, 'sbatch', '*.sbatch'))
    input_files = []
    n_trials_list = []
    for sbatch_file in sbatch_files:
        with open(sbatch_file) as f:
            for line in f.readlines():
                match = re.search(r'--file (.*json)', line)
                if match:
                    input_files.append(os.path.abspath(match.group(1)))

                    n_trials = 100
                    match_trials = re.search(r'--trials (\d+)', line)
                    if match:
                        n_trials = int(match_trials.group(1))
                    n_trials_list.append(n_trials)
    for input_file, n_trials in zip(input_files, n_trials_list):
        batch_sim = read_input_json(input_file)
        batch_sim.load_results()
        print(batch_sim.label)
        wall_time = batch_sim.wall_time
        time_remaining = batch_sim.estimate_remaining_time(n_trials)
        if wall_time + time_remaining > 0:
            progress = wall_time/(wall_time + time_remaining)
        else:
            progress = 0
        print('    target n_trials =', n_trials)
        print(
            '    wall time =',
            str(datetime.timedelta(seconds=wall_time))
        )
        print(
            '    remaining time =',
            str(datetime.timedelta(seconds=time_remaining))
        )
        print(f'    progress = {progress*100:.2f}%')



from typing import Dict, Tuple, Optional, List
from types import MethodType
from copy import copy
import os
from abc import ABCMeta, abstractmethod
import numpy as np
import json
from scipy.sparse import csr_matrix, dok_matrix

import panqec
from panqec.bpauli import bs_prod, get_effective_error
from panqec import bsparse

os.environ['PANQEC_ROOT_DIR'] = os.path.dirname(panqec.__file__)

Operator = Dict[Tuple, str]  # Coordinate to pauli ('X', 'Y' or 'Z')


class StabilizerCode(metaclass=ABCMeta):
    """Abstract class for generic stabilizer codes (CSS or not)

    Any subclass should override the following four methods:
    - get_qubit_coordinates() to define all the coordinates in the lattice
    that contain qubits
    - get_stabilizer_coordinates() to define all the coordinates in the lattice
    that contain stabilizers
    - qubit_axis(location) to return the axis of a qubit at a given location
    (when qubit have an orientation in space, for instance when they are edges)

    Using only those methods, a StabilizerCode will then automatically create
    the corresponding parity-check matrix (in self.stabilizers) and can be used
    to make a visualization in the GUI or calculate thresholds.
    """

    X_AXIS = 0
    Y_AXIS = 1
    Z_AXIS = 2

    deformation_names: List[str] = []

    def __init__(
        self, L_x: int,
        L_y: Optional[int] = None,
        L_z: Optional[int] = None,
    ):
        """Constructor for the StabilizerCode class

        Parameters
        ----------
        L_x : int
            Dimension of the lattice in the x direction (or in all directions
            if L_y and L_z are not given)
        L_y: int, optional
            Dimension of the lattice in the y direction
        L_z: int, optional
            Dimension of the lattice in the z direction
        """

        if L_y is None:
            L_y = L_x
        if L_z is None and self.dimension == 3:
            L_z = L_x

        self.L_x = L_x
        self.L_y = L_y
        self.L_z = L_z

        self._size: Tuple
        if self.dimension == 2:
            self._size = (L_x, L_y)
        else:
            self._size = (L_x, L_y, L_z)

        self._qubit_coordinates: List = []
        self._stabilizer_coordinates: List[Tuple] = []

        self._qubit_index: Dict[Tuple, int] = {}
        self._stabilizer_index: Dict[Tuple, int] = {}

        self._stabilizer_matrix = bsparse.empty_row(2*self.n)
        self._Hx = bsparse.empty_row(self.n)
        self._Hz = bsparse.empty_row(self.n)
        self._logicals_x: Optional[np.ndarray] = None
        self._logicals_z: Optional[np.ndarray] = None
        self._is_css: Optional[bool] = None
        self._x_indices: Optional[np.ndarray] = None
        self._z_indices: Optional[np.ndarray] = None
        self._d: Optional[int] = None
        self._stabilizer_types: Optional[List[str]] = None
        self.is_deformed: bool = False
        self.deformation_name: Optional[str] = None
        self.deformation_kwargs: Optional[dict] = None

        self.colormap = {'red': '0xFF4B3E',
                         'blue': '0x48BEFF',
                         'green': '0x058C42',
                         'pink': '0xffbcbc',
                         'white': '0xf2f2fc',
                         'gold': '0xf1c232',
                         'coral': '0xFA824C',
                         'light-yellow': '0xFAFAC6',
                         'salmon': '0xe79e90',
                         'light-orange': '0xFA824C',
                         'orange': '0xfa7921'}

    @property
    @abstractmethod
    def dimension(self) -> int:
        """Dimension of the code (usually 2 or 3)"""

    @property
    @abstractmethod
    def label(self) -> str:
        """Label uniquely identifying a code, including its lattice dimensions
        Example: 'Toric 3D {Lx}x{Ly}x{Lz}'
        """

    @property
    def id(self) -> str:
        """Returns a string identifying the class (usually the code name)"""
        return self.__class__.__name__

    @property
    def params(self) -> dict:
        """List of class arguments (as a dictionary), that can be saved
        and reused to instantiate the same code"""
        return {
            'L_x': self.L_x,
            'L_y': self.L_y,
            'L_z': self.L_z,
        }

    @property
    def n(self) -> int:
        """Number of physical qubits"""
        return len(self.qubit_coordinates)

    @property
    def k(self) -> int:
        """Number of logical qubits"""
        return self.logicals_x.shape[0]

    @property
    def d(self) -> int:
        """Distance of the code"""
        if self._d is None:
            weights_z = np.sum(
                np.logical_or(
                    self.logicals_z[:, :self.n],
                    self.logicals_z[:, self.n:]
                ),
                axis=1
            )
            weights_x = np.sum(
                np.logical_or(
                    self.logicals_x[:, :self.n],
                    self.logicals_x[:, self.n:]
                ), axis=1
            )

            self._d = min(np.min(weights_x), np.min(weights_z))

        return self._d

    @property
    def qubit_coordinates(self) -> List[Tuple]:
        """List of all the coordinates that contain a qubit"""

        if len(self._qubit_coordinates) == 0:
            self._qubit_coordinates = self.get_qubit_coordinates()

        return self._qubit_coordinates

    @property
    def stabilizer_coordinates(self) -> List[Tuple]:
        """List of all the coordinates that contain a stabilizer"""

        if len(self._stabilizer_coordinates) == 0:
            self._stabilizer_coordinates = self.get_stabilizer_coordinates()

        return self._stabilizer_coordinates

    @property
    def qubit_index(self) -> Dict[Tuple, int]:
        """Dictionary that assigns an index to a given qubit location"""

        if len(self._qubit_index) == 0:
            self._qubit_index = {
                loc: i for i, loc in enumerate(self.qubit_coordinates)
            }

        return self._qubit_index

    @property
    def stabilizer_index(self) -> Dict[Tuple, int]:
        """Dictionary that assigns an index to a given stabilizer location"""

        if len(self._stabilizer_index) == 0:
            self._stabilizer_index = {
                loc: i for i, loc in enumerate(self.stabilizer_coordinates)
            }

        return self._stabilizer_index

    @property
    def n_stabilizers(self) -> int:
        """Number of stabilizer generators"""
        return len(self.stabilizer_index)

    @property
    def logicals_x(self) -> np.ndarray:
        """Logical X operator, as a k x 2n matrix in the binary
        symplectic format, where k is the number of logical X operators,
        and n the number of qubits.
        """
        if self._logicals_x is None:
            logical_ops = self.get_logicals_x()
            k = len(logical_ops)
            self._logicals_x = np.zeros((k, 2*self.n), dtype='uint8')

            for i, logical_op in enumerate(logical_ops):
                self._logicals_x[i] = self.to_bsf(logical_op)

        return self._logicals_x

    @property
    def logicals_z(self) -> np.ndarray:
        """Logical Z operator, as a k x 2n matrix in the binary
        symplectic format, where k is the number of logical Z operators,
        and n the number of qubits.
        """
        if self._logicals_z is None:
            logical_ops = self.get_logicals_z()
            k = len(logical_ops)
            self._logicals_z = np.zeros((k, 2*self.n), dtype='uint8')

            for i, logical_op in enumerate(logical_ops):
                self._logicals_z[i] = self.to_bsf(logical_op)

        return self._logicals_z

    @property
    def is_css(self) -> bool:
        """Determines if a code is CSS, i.e. if it has separate X
        and Z stabilizers
        """
        if self._is_css is None:
            self._is_css = not np.any(
                np.logical_and(self.x_indices, self.z_indices)
            )

        return self._is_css

    @property
    def stabilizer_matrix(self) -> csr_matrix:
        """Parity-check matrix of the code in the binary symplectic format.
        It is a sparse matrix of dimension k x 2n, where k is the total number
        of stabilizers and n the number of qubits
        """

        if bsparse.is_empty(self._stabilizer_matrix):
            sparse_dict: Dict = dict()
            self._stabilizer_matrix = dok_matrix(
                (self.n_stabilizers, 2*self.n),
                dtype='uint8'
            )

            for i_stab, stabilizer_location in enumerate(
                self.stabilizer_coordinates
            ):
                stabilizer_op = self.get_stabilizer(stabilizer_location)

                for qubit_location in stabilizer_op.keys():
                    if stabilizer_op[qubit_location] in ['X', 'Y']:
                        i_qubit = self.qubit_index[qubit_location]
                        if (i_stab, i_qubit) in sparse_dict.keys():
                            sparse_dict[(i_stab, i_qubit)] += 1
                        else:
                            sparse_dict[(i_stab, i_qubit)] = 1
                    if stabilizer_op[qubit_location] in ['Y', 'Z']:
                        i_qubit = self.n + self.qubit_index[qubit_location]
                        if (i_stab, i_qubit) in sparse_dict.keys():
                            sparse_dict[(i_stab, i_qubit)] += 1
                        else:
                            sparse_dict[(i_stab, i_qubit)] = 1

            self._stabilizer_matrix._update(sparse_dict)
            self._stabilizer_matrix = self._stabilizer_matrix.tocsr()
            self._stabilizer_matrix.data %= 2

        return self._stabilizer_matrix

    @property
    def size(self) -> Tuple:
        """Dimensions of the lattice."""
        return self._size

    @property
    def Hx(self) -> csr_matrix:
        """Parity-check matrix corresponding to the X stabilizers of the code.
        It is a sparse matrix of dimension k x n, where k is the number of
        X stabilizers and n the number of qubits.
        Works only for CSS codes.
        """

        if not self.is_css:
            raise ValueError("Impossible to extract Hz: the code is not CSS")

        if self._Hx.shape[0] == 0:
            H = self.stabilizer_matrix[:, :self.n]
            self._Hx = H[self.x_indices]

        return self._Hx

    @property
    def Hz(self) -> csr_matrix:
        """Parity-check matrix corresponding to the Z stabilizers of the code.
        It is a sparse matrix of dimension k x n, where k is the number of
        Z stabilizers and n the number of qubits.
        Works only for CSS codes.
        """

        if not self.is_css:
            raise ValueError("Impossible to extract Hz: the code is not CSS")

        if self._Hz.shape[0] == 0:
            H = self.stabilizer_matrix[:, self.n:]
            self._Hz = H[self.z_indices]

        return self._Hz

    @property
    def x_indices(self) -> np.ndarray:
        """Indices of the X stabilizers in the parity-check matrix,
        as a boolean array s.t. x_indices[i] is True if stabilizer H[i]
        only contain X operators and False otherwise"""

        if self._x_indices is None:
            Hx = self.stabilizer_matrix[:, :self.n]
            self._x_indices = (Hx.getnnz(1) > 0)

        return self._x_indices

    @property
    def z_indices(self) -> np.ndarray:
        """Indices of the Z stabilizers in the parity-check matrix,
        as a boolean array s.t. z_indices[i] is True if stabilizer H[i]
        only contain Z operators and False otherwise"""

        if self._z_indices is None:
            Hz = self.stabilizer_matrix[:, self.n:]
            self._z_indices = (Hz.getnnz(1) > 0)

        return self._z_indices

    def in_codespace(self, error: np.ndarray) -> bool:
        """Check whether or not a given error is in the codespace,
        i.e. whether it has a zero syndrome or not.

        Parameters
        ----------
        error: np.ndarray
            Error as an array of size 2n (where n is the number of qubits)
            in the binary symplectic format

        Returns
        -------
        bool
            Whether or not the error is in the codespace
        """

        return bool(np.all(self.measure_syndrome(error) == 0))

    def logical_errors(self, error: np.ndarray) -> np.ndarray:
        """Return the logical errors, as an array of size 2k
        (where k is the number of logicals), such that each component is
        1 if and only if it anticommutes with the corresponding logical.

        By convention, the first k indices correspond to the X logicals
        and the last k to the the Z logicals

        Parameters
        ----------
        error: np.ndarray
            Error as an array of size 2n (where n is the number of qubits)
            in the binary symplectic format

        Returns
        -------
        logical_errors: np.ndarray
            Array of size 2k (where k is the number of logicals)
            indicating whether the error commute with each X and Z logical.
        """

        return get_effective_error(
            error, self.logicals_x, self.logicals_z
        )

    def is_logical_error(self, error) -> bool:
        """Check whether or not a given error is in the codespace,
        i.e. whether it has a zero syndrome or not.

        Parameters
        ----------
        error: np.ndarray
            Error as an array of size 2n (where n is the number of qubits)
            in the binary symplectic format

        Returns
        -------
        bool
            Whether or not the error is in the codespace
        """

        return bool(np.any(self.logical_errors(error) != 0))

    def is_success(self, total_error) -> bool:
        return (self.in_codespace(total_error) and
                not self.is_logical_error(total_error))

    def extract_x_syndrome(self, syndrome: np.ndarray) -> np.ndarray:
        """For CSS codes only. Returns the part of the syndrome that
        corresponds to X stabilizers.

        Parameters
        ----------
        syndrome: np.ndarray
            Syndrome as a sparse row of dimension 1xm, where m is the number
            of stabilizers.

        Returns
        -------
        x_syndrome: np.ndarray
            Syndrome reduced to X stabilizers
        """

        return syndrome[self.x_indices]

    def extract_z_syndrome(self, syndrome: np.ndarray) -> np.ndarray:
        """For CSS codes only. Returns the part of the syndrome that
        corresponds to Z stabilizers.

        Parameters
        ----------
        syndrome: np.ndarray
            Syndrome as a sparse row of dimension 1xm, where m is the number
            of stabilizers.

        Returns
        -------
        z_syndrome: np.ndarray
            Syndrome reduced to X stabilizers
        """

        return syndrome[self.z_indices]

    def to_bsf(self, operator: Operator) -> np.ndarray:
        """Convert an operator (given as a dictionary qubit_location -> pauli)
        to an array in the binary symplectic format.

        Parameters
        ----------
        operator: Dict[Tuple, str]
            Operator given as a dictionary that assigns a Pauli operator
            ('X', 'Y' or 'Z') to each qubit location in its support

        Returns
        -------
        bsf_operator: np.ndarray
            Array of dimension 2n in the binary symplectic format
            (where n is the number of qubits)
        """
        bsf_operator = np.zeros(2*self.n, dtype=np.uint)

        for qubit_location in operator.keys():
            if operator[qubit_location] in ['X', 'Y']:
                bsf_operator[self.qubit_index[qubit_location]] += 1
            if operator[qubit_location] in ['Y', 'Z']:
                bsf_operator[self.n + self.qubit_index[qubit_location]] += 1

        return bsf_operator

    def from_bsf(self, bsf_operator: np.ndarray) -> Operator:
        """Convert an operator given as a sparse row in the binary
        symplectic format to a dictionary qubit_location -> pauli.

        Parameters
        ----------
        bsf_operator: np.ndarray
            Array of dimension (1, 2n) in the binary symplectic format
            (where n is the number of qubits)
        Returns
        -------
        operator: Dict[Tuple, str]
            Operator given as a dictionary that assigns a Pauli operator
            ('X', 'Y' or 'Z') to each qubit location in its support

        """
        assert (
            bsf_operator.shape[0] == 1 or len(bsf_operator.shape) == 1
        ), "Can only take one operator at a time."

        operator = dict()

        if len(bsf_operator.shape) == 1:
            cols = bsf_operator.nonzero()[0]
        else:
            rows, cols = bsf_operator.nonzero()

        for col in cols:
            if col < self.n:
                location = self.qubit_coordinates[col]
                operator[location] = 'X'
            else:
                location = self.qubit_coordinates[col - self.n]
                if location in operator.keys():
                    operator[location] = 'Y'
                else:
                    operator[location] = 'Z'

        return operator

    def measure_syndrome(self, error: np.ndarray) -> np.ndarray:
        """Noiseless syndrome corresponding to a given Pauli error.

        Parameters
        ----------
        error: np.ndarray
            Error given as an array of dimension 2n in the binary
            symplectic format.

        Returns
        -------
        syndrome: np.ndarray
            Syndrome, as an array of dimension m (where m is the number
            of stabilizers)
        """

        return bs_prod(self.stabilizer_matrix, error)

    def is_stabilizer(self, location: Tuple, stab_type: Optional[str] = None):
        """Returns whether a given location in the coordinate system
        corresponds to a stabilizer or not
        """
        _is_stabilizer = (
            (location in self.stabilizer_index) and
            (stab_type is None or self.stabilizer_type(location) == stab_type)
        )

        return _is_stabilizer

    def is_qubit(self, location: Tuple):
        """Returns whether a given location in the coordinate system
        corresponds to a qubit or not. It is done by checking that the input
        location is a key in the dictionary `self.qubit_index`.

        Parameters
        ----------
        location : Tuple
            Location as a tuple of coordinates

        Returns
        -------
        Bool
            Whether the location is a qubit in the coordinate system.
        """
        return location in self.qubit_index

    @abstractmethod
    def get_qubit_coordinates(self) -> List[Tuple]:
        """Give the list of all the qubit coordinates, in a coordinate system
        that should contain both the qubits and the stabilizers.
        This function is used to set the attributes `self.qubit_coordinates`
        and `self.qubit_index`.

        Returns
        -------
        qubit_coordinates: List[Tuple]
            List of coordinates
        """

    @abstractmethod
    def get_stabilizer_coordinates(self) -> List[Tuple]:
        """Create list of stabilizer coordinates, in a coordinate system
        that should contain both the qubits and the stabilizers.
        This function is used to set the attributes
        `self.stabilizer_coordinates` and `self.stabilizer_index`.
        """

    @abstractmethod
    def qubit_axis(self, location: Tuple) -> str:
        """ Return the orientation of a qubit sitting at given location
        (as a string representing the axis 'x', 'y' or 'z').
        Useful when qubits have an orientation in space, for instance when
        they are edges, to help establish the visual representation of the
        code in the GUI, to simplify the construction of stabilizers,
        and to create Clifford deformations.

        Parameters
        ----------
        location: Tuple
            Location of the qubit in the coordinate system.

        Returns
        -------
        axis: str
            Either 'x', 'y' or 'z', depending on the orientation axis of the
            qubit.
        """

    @abstractmethod
    def stabilizer_type(self, location: Tuple) -> str:
        """ Returns the type of a stabilizer sitting at a given location.
        E.g. 'vertex' or 'face' in toric codes
        """

    @abstractmethod
    def get_stabilizer(self, location: Tuple) -> Operator:
        """ Returns a stabilizer, formatted as dictionary that assigns a Pauli
        operator ('X', 'Y' or 'Z') to each qubit location in the support of
        the stabilizer.

        For example, for a vertex stabilizer in the 2D toric code, we could
        have
        `get_stabilizer((1,1)) -> {(1,0): 'X', (0, 1): 'X', (2, 1): 'X',
        (1, 2): 'X'}`

        Parameters
        ----------
        location: Tuple
            Location of the stabilizer in the coordinate system

        Returns
        -------
        stabilizer: Dict[Tuple, str]
            Dictionary that assigns a Pauli operator ('X', 'Y' or 'Z') to each
            qubit location in the support of the stabilizer
        """

    @abstractmethod
    def get_logicals_x(self) -> List[Operator]:
        """Returns the list of logical X operators, where each operator is a
        dictionary that assigns a Pauli operator ('X', 'Y' or 'Z') to each
        qubit location in its support.

        Returns
        -------
        logicals: List[Dict[Tuple, str]]
            List of dictionaries, where each dictionary assign a Pauli
            operator ('X', 'Y' or 'Z') to each qubit location in the support
            of the logical operator.
        """

    @abstractmethod
    def get_logicals_z(self) -> List[Operator]:
        """Returns the list of logical Z operators, where each operator is a
        dictionary that assigns a Pauli operator ('X', 'Y' or 'Z') to each
        qubit location in its support.

        Returns
        -------
        logicals: List[Dict[Tuple, str]]
            List of dictionaries, where each dictionary assign a Pauli
            operator ('X', 'Y' or 'Z') to each qubit location in the support
            of the logical operator.
        """

    def get_deformation(
        self, location: Tuple, deformation_name: str, **kwargs
    ):
        return NotImplementedError("No deformation implemented for this code")

    def deform(self, deformation_name, **kwargs):
        self.__init__(*self.size)

        self.is_deformed = True
        self.deformation_name = deformation_name
        self.deformation_kwargs = kwargs

        if not hasattr(self, '_get_undeformed_stabilizer'):
            self._get_undeformed_stabilizer = copy(
                MethodType(self.get_stabilizer, self)
            )

        if not hasattr(self, '_get_undeformed_logicals_x'):
            self._get_undeformed_logicals_x = copy(
                MethodType(self.get_logicals_x, self)
            )

        if not hasattr(self, '_get_undeformed_logicals_z'):
            self._get_undeformed_logicals_z = copy(
                MethodType(self.get_logicals_z, self)
            )

        def get_stabilizer(self, location):
            stab = self._get_undeformed_stabilizer(location)

            for loc in stab.keys():
                deformation = self.get_deformation(
                    loc, deformation_name, **kwargs
                )
                stab[loc] = deformation[stab[loc]]

            return stab

        def get_logicals_x(self):
            logicals = self._get_undeformed_logicals_x()

            for logical in logicals:
                for loc in logical.keys():
                    deformation = self.get_deformation(
                        loc, deformation_name, **kwargs
                    )
                    logical[loc] = deformation[logical[loc]]

            return logicals

        def get_logicals_z(self):
            logicals = self._get_undeformed_logicals_z()

            for logical in logicals:
                for loc in logical.keys():
                    deformation = self.get_deformation(
                        loc, deformation_name, **kwargs
                    )
                    logical[loc] = deformation[logical[loc]]

            return logicals

        self.get_stabilizer = MethodType(get_stabilizer, self)
        self.get_logicals_x = MethodType(get_logicals_x, self)
        self.get_logicals_z = MethodType(get_logicals_z, self)

    def stabilizer_representation(self,
                                  location: Tuple,
                                  rotated_picture=False,
                                  json_file=None) -> Dict:
        """Returns a dictionary of visualization parameters for the input
        stabilizer, that can be used by the web visualizer.

        It should contain 4 keys:
        - 'type': the type of stabilizer, e.g. 'vertex'
        - 'location': [x, y, z],
        - 'object': the type of object to use for visualization, e.g. 'sphere'
        - 'params': a dictionary of parameters for the chosen object

        Parameters
        ----------
        location: Tuple
            Coordinates of the stabilizer
        rotated_picture: bool
            For codes that have a rotated picture, can be used to differentiate
            the two types visualizations
        json_file: str
            File with the initial configuration for the code

        Returns
        -------
        representation: Dict
            Dictionary to send to the GUI
        """
        if json_file is None:
            json_file = os.path.join(
                os.environ['PANQEC_ROOT_DIR'], 'codes', 'gui-config.json'
            )

        stab_type = self.stabilizer_type(location)

        with open(json_file, 'r') as f:
            data = json.load(f)

        code_name = self.id
        picture = 'rotated' if rotated_picture else 'kitaev'

        representation = data[code_name]['stabilizers'][picture][stab_type]
        representation['type'] = stab_type
        representation['location'] = location

        for activation in ['activated', 'deactivated']:
            color_name = representation['color'][activation]
            representation['color'][activation] = self.colormap[color_name]

        return representation

    def qubit_representation(self,
                             location: Tuple,
                             rotated_picture=False,
                             json_file=None) -> Dict:
        """Returns a dictionary of visualization parameters for the input
        qubit,  that can be used by the web visualizer.
        - 'location': [x, y, z],
        - 'object': the type of object to use for visualization, e.g. 'sphere'
        - 'params': a dictionary of parameters for the chosen object

        Parameters
        ----------
        location: Tuple
            Coordinates of the qubit
        rotated_picture: bool
            For codes that have a rotated picture, can be used to differentiate
            the two types visualizations
        json_file: str
            File with the initial configuration for the code

        Returns
        -------
        representation: Dict
            Dictionary to send to the GUI
        """
        if json_file is None:
            json_file = os.path.join(
                os.environ['PANQEC_ROOT_DIR'], 'codes', 'gui-config.json'
            )

        with open(json_file, 'r') as f:
            data = json.load(f)

        code_name = self.id

        picture = 'rotated' if rotated_picture else 'kitaev'

        representation = data[code_name]['qubits'][picture]

        representation['params']['axis'] = self.qubit_axis(location)
        representation['location'] = location

        for pauli in ['I', 'X', 'Y', 'Z']:
            color_name = representation['color'][pauli]
            representation['color'][pauli] = self.colormap[color_name]

        return representation

    def type_index(self, stab_type: str) -> Dict[Tuple, int]:
        """Dictionary of locations and indices for given stabilizer type.

        Parameters
        ----------
        stab_type: str
            Stabilizer type ot index.

        Returns
        -------
        index: Dict[Tuple, int]
            Dictionary of qubit indices for each stabilizer location that
            matches the given type.
        """
        return {
            location: index
            for index, location in enumerate(self.stabilizer_coordinates)
            if self.stabilizer_type(location) == stab_type
        }

    @property
    def stabilizer_types(self):
        if self._stabilizer_types is None:
            self._stabilizer_types = list(set(
                self.stabilizer_type(location)
                for location in self.stabilizer_coordinates
            ))
        return self._stabilizer_types

    def site(self, operator: Operator, pauli: str, location: Tuple) -> None:
        """Apply a Pauli on operator at site location.

        Note that the operator is a (mutable) dict.

        Parameters
        ----------
        operator: Operator
            Operator in dictionary representation.
        pauli: str
            Pauli to apply.
        """

        product_map = {
            ('X', 'Y'): 'Z',
            ('X', 'Z'): 'Y',
            ('Y', 'X'): 'Z',
            ('Y', 'Z'): 'X',
            ('Z', 'X'): 'Y',
            ('Z', 'Y'): 'X',
        }

        if location in operator:
            if operator[location] == pauli:
                operator.pop(location)
            else:
                operator[location] = product_map[(operator[location], pauli)]
        else:
            operator[location] = pauli



from typing import Tuple, Dict, List
from panqec.codes import StabilizerCode

Operator = Dict[Tuple, str]  # Location to pauli ('X','Y','Z')
Coordinates = List[Tuple]  # List of locations


class Toric3DCode(StabilizerCode):
    """3D surface code on periodic cubic lattice with qubits on edges.

    Parameters
    ----------
    L_x : int
        Size in the x direction.
    L_y : Optional[int]
        Size in the y direction, assumed same as Lx if not given.
    L_z : Optional[int]
        Size in the z direction, assumed same as Lx if not given.

    Notes
    -----
    The qubits live on the edges of the 3D lattice.
    There are two types of stabilizer generators:
    6-body vertex operators living on vertices,
    and 4-body face operators living on faces.

    In the coordinate system used in this implementation,
    the origin (0, 0, 0) is a vertex,
    and each unit cell is a cube of linear size 2.
    """
    dimension = 3
    deformation_names = ['XZZX']

    @property
    def label(self) -> str:
        return 'Toric {}x{}x{}'.format(*self.size)

    def get_qubit_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly, Lz = self.size

        # Qubits along e_x
        for x in range(1, 2*Lx, 2):
            for y in range(0, 2*Ly, 2):
                for z in range(0, 2*Lz, 2):
                    coordinates.append((x, y, z))

        # Qubits along e_y
        for x in range(0, 2*Lx, 2):
            for y in range(1, 2*Ly, 2):
                for z in range(0, 2*Lz, 2):
                    coordinates.append((x, y, z))

        # Qubits along e_z
        for x in range(0, 2*Lx, 2):
            for y in range(0, 2*Ly, 2):
                for z in range(1, 2*Lz, 2):
                    coordinates.append((x, y, z))

        return coordinates

    def get_stabilizer_coordinates(self) -> Coordinates:
        coordinates: Coordinates = []
        Lx, Ly, Lz = self.size

        # Vertices
        for x in range(0, 2*Lx, 2):
            for y in range(0, 2*Ly, 2):
                for z in range(0, 2*Lz, 2):
                    coordinates.append((x, y, z))

        # Face in xy plane
        for x in range(1, 2*Lx, 2):
            for y in range(1, 2*Ly, 2):
                for z in range(0, 2*Lz, 2):
                    coordinates.append((x, y, z))

        # Face in yz plane
        for x in range(0, 2*Lx, 2):
            for y in range(1, 2*Ly, 2):
                for z in range(1, 2*Lz, 2):
                    coordinates.append((x, y, z))

        # Face in xz plane
        for x in range(1, 2*Lx, 2):
            for y in range(0, 2*Ly, 2):
                for z in range(1, 2*Lz, 2):
                    coordinates.append((x, y, z))

        return coordinates

    def stabilizer_type(self, location: Tuple) -> str:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location} for a stabilizer")

        x, y, z = location
        if x % 2 == 0 and y % 2 == 0:
            return 'vertex'
        else:
            return 'face'

    def get_stabilizer(self, location) -> Operator:
        if not self.is_stabilizer(location):
            raise ValueError(f"Invalid coordinate {location} for a stabilizer")

        if self.stabilizer_type(location) == 'vertex':
            pauli = 'Z'
        else:
            pauli = 'X'

        x, y, z = location

        if self.stabilizer_type(location) == 'vertex':
            delta = [(-1, 0, 0), (1, 0, 0), (0, -1, 0), (0, 1, 0), (0, 0, -1),
                     (0, 0, 1)]
        else:
            # Face in xy-plane.
            if z % 2 == 0:
                delta = [(-1, 0, 0), (1, 0, 0), (0, -1, 0), (0, 1, 0)]
            # Face in yz-plane.
            elif (x % 2 == 0):
                delta = [(0, -1, 0), (0, 1, 0), (0, 0, -1), (0, 0, 1)]
            # Face in zx-plane.
            elif (y % 2 == 0):
                delta = [(-1, 0, 0), (1, 0, 0), (0, 0, -1), (0, 0, 1)]

        operator: Operator = dict()
        for d in delta:
            Lx, Ly, Lz = self.size
            qubit_location = ((x + d[0]) % (2*Lx), (y + d[1]) % (2*Ly),
                              (z + d[2]) % (2*Lz))

            if self.is_qubit(qubit_location):
                operator[qubit_location] = pauli

        return operator

    def qubit_axis(self, location):
        x, y, z = location

        if (z % 2 == 0) and (x % 2 == 1) and (y % 2 == 0):
            axis = 'x'
        elif (z % 2 == 0) and (x % 2 == 0) and (y % 2 == 1):
            axis = 'y'
        elif (z % 2 == 1) and (x % 2 == 0) and (y % 2 == 0):
            axis = 'z'
        else:
            raise ValueError(f"Location {location} does not correspond"
                             "to a qubit")

        return axis

    def get_logicals_x(self) -> List[Operator]:
        """The 3 logical X operators."""

        Lx, Ly, Lz = self.size
        logicals = []

        # X operators along x edges in x direction.
        operator: Operator = dict()
        for x in range(1, 2*Lx, 2):
            operator[(x, 0, 0)] = 'X'
        logicals.append(operator)

        # X operators along y edges in y direction.
        operator = dict()
        for y in range(1, 2*Ly, 2):
            operator[(0, y, 0)] = 'X'
        logicals.append(operator)

        # X operators along z edges in z direction
        operator = dict()
        for z in range(1, 2*Lz, 2):
            operator[(0, 0, z)] = 'X'
        logicals.append(operator)

        return logicals

    def get_logicals_z(self) -> List[Operator]:
        """Get the 3 logical Z operators."""
        Lx, Ly, Lz = self.size
        logicals = []

        # Z operators on x edges forming surface normal to x (yz plane).
        operator: Operator = dict()
        for y in range(0, 2*Ly, 2):
            for z in range(0, 2*Lz, 2):
                operator[(1, y, z)] = 'Z'
        logicals.append(operator)

        # Z operators on y edges forming surface normal to y (zx plane).
        operator = dict()
        for z in range(0, 2*Lz, 2):
            for x in range(0, 2*Lx, 2):
                operator[(x, 1, z)] = 'Z'
        logicals.append(operator)

        # Z operators on z edges forming surface normal to z (xy plane).
        operator = dict()
        for x in range(0, 2*Lx, 2):
            for y in range(0, 2*Ly, 2):
                operator[(x, y, 1)] = 'Z'
        logicals.append(operator)

        return logicals

    def stabilizer_representation(
        self, location, rotated_picture=False, json_file=None
    ) -> Dict:
        representation = super().stabilizer_representation(
            location, rotated_picture, json_file=json_file
        )

        x, y, z = location
        if not rotated_picture and self.stabilizer_type(location) == 'face':
            if z % 2 == 0:  # xy plane
                representation['params']['normal'] = [0, 0, 1]
            elif x % 2 == 0:  # yz plane
                representation['params']['normal'] = [1, 0, 0]
            else:  # xz plane
                representation['params']['normal'] = [0, 1, 0]

        if rotated_picture and self.stabilizer_type(location) == 'face':
            if z % 2 == 0:
                representation['params']['normal'] = [0, 0, 1]
            elif x % 2 == 0:
                representation['params']['normal'] = [1, 0, 0]
            else:
                representation['params']['normal'] = [0, 1, 0]

        return representation

    def get_deformation(
        self, location: Tuple,
        deformation_name: str,
        deformation_axis: str = 'y',
        **kwargs
    ) -> Dict:

        if deformation_axis not in ['x', 'y', 'z']:
            raise ValueError(f"{deformation_axis} is not a valid "
                             "deformation axis")

        if deformation_name == 'XZZX':
            undeformed_dict = {'X': 'X', 'Y': 'Y', 'Z': 'Z'}
            deformed_dict = {'X': 'Z', 'Y': 'Y', 'Z': 'X'}

            if self.qubit_axis(location) == deformation_axis:
                deformation = deformed_dict
            else:
                deformation = undeformed_dict

        else:
            raise ValueError(f"The deformation {deformation_name}"
                             "does not exist")

        return deformation



from importlib import reload
import pytest
import os
import panqec
from panqec.config import PANQEC_DIR, PANQEC_DARK_THEME


def test_output_dir_exists():
    assert os.path.exists(PANQEC_DIR)


def test_error_raised_if_panqec_dir_does_not_exist(monkeypatch):
    monkeypatch.setenv('PANQEC_DIR', '/fake/path/here/')
    with pytest.raises(FileNotFoundError):
        reload(panqec.config)


def test_dark_theme_known():
    assert type(PANQEC_DARK_THEME) is bool



import pytest
import numpy as np
from panqec.bpauli import bsf_wt
from panqec.codes import Toric3DCode


@pytest.fixture
def code():
    return Toric3DCode(9, 10, 11)


class TestToric3DPauli:

    def test_x_block_z_block_to_bsf_and_back(self, code):
        L_x, L_y, L_z = code.size

        # Deterministic example of X and Z blocks in 4D array.
        np.random.seed(0)
        x_block = np.random.randint(0, 2, size=3*L_x*L_y*L_z)
        z_block = np.random.randint(0, 2, size=3*L_x*L_y*L_z)

        # Ensure non-trivial.
        assert np.any(z_block == 1)
        assert np.any(x_block == 1)
        assert np.any(z_block != x_block)

        # Construc the bsf.
        bsf = np.hstack([x_block, z_block])
        operator = code.from_bsf(bsf)

        # Convert to binary simplectic form.
        new_bsf = code.to_bsf(operator)
        assert np.all(new_bsf == bsf)

    def test_apply_operator_on_site(self, code):
        operator = {
            (0, 0, 1): 'X'
        }
        assert bsf_wt(code.to_bsf(operator)) == 1

    def test_apply_operators_on_vertex(self, code):
        location = (2, 2, 2)
        assert code.stabilizer_type(location) == 'vertex'
        vertex_operator = code.get_stabilizer(location)

        # Each vertex has 6 edges.
        assert bsf_wt(code.to_bsf(vertex_operator)) == 6

        # Same operator applied site by site.
        operator = dict()
        operator[(2, 3, 2)] = 'Z'
        operator[(3, 2, 2)] = 'Z'
        operator[(1, 2, 2)] = 'Z'
        operator[(2, 1, 2)] = 'Z'
        operator[(2, 2, 1)] = 'Z'
        operator[(2, 2, 3)] = 'Z'

        assert bsf_wt(code.to_bsf(operator)) == 6

        from panqec.utils import simple_print
        simple_print(code.to_bsf(vertex_operator))
        simple_print(code.to_bsf(operator))
        assert np.all(
            code.to_bsf(vertex_operator) == code.to_bsf(operator)
        )

        assert vertex_operator == operator

    @pytest.mark.parametrize('location', [(0, 1, 1), (1, 0, 1), (1, 1, 0)])
    def test_apply_face_operators_on_face(self, code, location):
        assert code.stabilizer_type(location) == 'face'
        face_operator = code.get_stabilizer(location)

        # Each face has 4 edges.
        assert bsf_wt(code.to_bsf(face_operator)) == 4

    def test_make_a_2x2_loop(self, code):
        faces = [(0, 1, 1), (0, 3, 1), (0, 1, 3), (0, 3, 3)]
        assert all(code.stabilizer_type(face) == 'face' for face in faces)
        loop = code.from_bsf(sum(
            code.to_bsf(code.get_stabilizer(face))
            for face in faces
        ) % 2)
        assert bsf_wt(code.to_bsf(loop)) == 8
        assert loop == {
            (0, 1, 0): 'X', (0, 3, 0): 'X',
            (0, 4, 1): 'X', (0, 4, 3): 'X',
            (0, 3, 4): 'X', (0, 1, 4): 'X',
            (0, 0, 3): 'X', (0, 0, 1): 'X',
        }

    def test_make_a_tetris_T_loop(self, code):
        """Product of faces like this should have weight 10.

        +-Z-+-Z-+-Z-+
        Z   |   |   Z
        +-Z-+---+-Z-+
            Z   Z
            +-Z-+

        """
        faces = [
            (3, 2, 1), (3, 2, 3), (1, 2, 3), (5, 2, 3),
        ]
        loop = code.from_bsf(sum(
            code.to_bsf(code.get_stabilizer(face))
            for face in faces
        ) % 2)
        assert bsf_wt(code.to_bsf(loop)) == 10

    def test_faces_sharing_same_vertex(self, code):
        r"""This thing should have weight six.

             .---Y---.            Coordinate axes:
            /       / \                      y
           Y       /   Y                    /
          /       /     \                  /
         .-------o       .         z <----o
          \       \     /                  \
           Y       \   Y                    \
            \       \ /                      x
             .---Y---.

        """
        faces = [(1, 0, 1), (0, 1, 1), (1, 1, 2)]
        loop = code.from_bsf(sum(
            code.to_bsf(code.get_stabilizer(face))
            for face in faces
        ) % 2)

        assert bsf_wt(code.to_bsf(loop)) == 6



import numpy as np
from ldpc import bposd_decoder
from panqec.codes import StabilizerCode
from panqec.error_models import BaseErrorModel
from panqec.decoders import BaseDecoder


class BeliefPropagationOSDDecoder(BaseDecoder):
    label = 'BP-OSD decoder'
    allowed_codes = None  # all codes allowed

    def __init__(self,
                 code: StabilizerCode,
                 error_model: BaseErrorModel,
                 error_rate: float,
                 max_bp_iter: int = 1000,
                 channel_update: bool = False,
                 osd_order: int = 10,
                 bp_method: str = 'msl'):
        super().__init__(code, error_model, error_rate)
        self._max_bp_iter = max_bp_iter
        self._channel_update = channel_update
        self._osd_order = osd_order
        self._bp_method = bp_method

        # Do not initialize the decoder until we call the decode method.
        # This is required because during analysis, there is no need to
        # initialize the decoder every time.
        self._initialized = False

    @property
    def params(self) -> dict:
        return {
            'max_bp_iter': self._max_bp_iter,
            'channel_update': self._channel_update,
            'osd_order': self._osd_order,
            'bp_method': self._bp_method
        }

    def get_probabilities(self):
        pi, px, py, pz = self.error_model.probability_distribution(
            self.code, self.error_rate
        )

        return pi, px, py, pz

    def update_probabilities(self, correction: np.ndarray,
                             px: np.ndarray, py: np.ndarray, pz: np.ndarray,
                             direction: str = "x->z") -> np.ndarray:
        """Update X probabilities once a Z correction has been applied"""

        n_qubits = correction.shape[0]

        new_probs = np.zeros(n_qubits)

        if direction == "z->x":
            for i in range(n_qubits):
                if correction[i] == 1:
                    if pz[i] + py[i] != 0:
                        new_probs[i] = py[i] / (pz[i] + py[i])
                else:
                    new_probs[i] = px[i] / (1 - pz[i] - py[i])

        elif direction == "x->z":
            for i in range(n_qubits):
                if correction[i] == 1:
                    if px[i] + py[i] != 0:
                        new_probs[i] = py[i] / (px[i] + py[i])
                else:
                    new_probs[i] = pz[i] / (1 - px[i] - py[i])

        else:
            raise ValueError(
                f"Unrecognized direction {direction} when "
                "updating probabilities"
            )

        return new_probs

    def initialize_decoders(self):
        is_css = self.code.is_css

        if is_css:
            self.z_decoder = bposd_decoder(
                self.code.Hx,
                error_rate=self.error_rate,
                max_iter=self._max_bp_iter,
                bp_method=self._bp_method,
                ms_scaling_factor=0,
                osd_method="osd_cs",  # Choose from: "osd_e", "osd_cs", "osd0"
                osd_order=self._osd_order
            )

            self.x_decoder = bposd_decoder(
                self.code.Hz,
                error_rate=self.error_rate,
                max_iter=self._max_bp_iter,
                bp_method=self._bp_method,
                ms_scaling_factor=0,
                osd_method="osd_cs",  # Choose from: "osd_e", "osd_cs", "osd0"
                osd_order=self._osd_order
            )

        else:
            self.decoder = bposd_decoder(
                self.code.stabilizer_matrix,
                error_rate=self.error_rate,
                max_iter=self._max_bp_iter,
                bp_method=self._bp_method,
                ms_scaling_factor=0,
                osd_method="osd_cs",  # Choose from: "osd_e", "osd_cs", "osd0"
                osd_order=self._osd_order
            )
        self._initialized = True

    def decode(self, syndrome: np.ndarray, **kwargs) -> np.ndarray:
        """Get X and Z corrections given code and measured syndrome."""

        if not self._initialized:
            self.initialize_decoders()

        is_css = self.code.is_css
        n_qubits = self.code.n
        syndrome = np.array(syndrome, dtype=int)

        if is_css:
            syndrome_z = self.code.extract_z_syndrome(syndrome)
            syndrome_x = self.code.extract_x_syndrome(syndrome)

        pi, px, py, pz = self.get_probabilities()

        probabilities_x = px + py
        probabilities_z = pz + py

        probabilities = np.hstack([probabilities_z, probabilities_x])

        if is_css:
            # Update probabilities (in case the distribution is new at each
            # iteration)
            self.x_decoder.update_channel_probs(probabilities_x)
            self.z_decoder.update_channel_probs(probabilities_z)

            # Decode Z errors
            self.z_decoder.decode(syndrome_x)
            z_correction = self.z_decoder.osdw_decoding

            # Bayes update of the probability
            if self._channel_update:
                new_x_probs = self.update_probabilities(
                    z_correction, px, py, pz, direction="z->x"
                )
                self.x_decoder.update_channel_probs(new_x_probs)

            # Decode X errors
            self.x_decoder.decode(syndrome_z)
            x_correction = self.x_decoder.osdw_decoding

            correction = np.concatenate([x_correction, z_correction])
        else:
            # Update probabilities (in case the distribution is new at each
            # iteration)
            self.decoder.update_channel_probs(probabilities)

            # Decode all errors
            self.decoder.decode(syndrome)
            correction = self.decoder.osdw_decoding
            correction = np.concatenate(
                [correction[n_qubits:], correction[:n_qubits]]
            )

        return correction


def test_decoder():
    from panqec.codes import XCubeCode
    from panqec.error_models import PauliErrorModel
    import time
    rng = np.random.default_rng()

    L = 20
    code = XCubeCode(L, L, L)

    error_rate = 0.5
    r_x, r_y, r_z = [0.15, 0.15, 0.7]
    error_model = PauliErrorModel(r_x, r_y, r_z)

    print("Create stabilizer matrix")
    code.stabilizer_matrix

    print("Create Hx and Hz")
    code.Hx
    code.Hz

    print("Create logicals")
    code.logicals_x
    code.logicals_z

    print("Instantiate BP-OSD")
    decoder = BeliefPropagationOSDDecoder(
        code, error_model, error_rate, osd_order=0, max_bp_iter=1000
    )

    # Start timer
    start = time.time()

    n_iter = 1
    accuracy = 0
    for i in range(n_iter):
        print(f"\nRun {code.label} {i}...")
        print("Generate errors")
        error = error_model.generate(code, error_rate, rng=rng)
        print("Calculate syndrome")
        syndrome = code.measure_syndrome(error)
        print("Decode")
        correction = decoder.decode(syndrome)
        print("Get total error")
        total_error = (correction + error) % 2

        codespace = code.in_codespace(total_error)
        success = not code.is_logical_error(total_error) and codespace
        print(success)
        accuracy += success

    accuracy /= n_iter
    print("Average time per iteration", (time.time() - start) / n_iter)
    print("Logical error rate", 1 - accuracy)


if __name__ == '__main__':
    test_decoder()



import panqec.bsparse as bsparse
import numpy as np
from scipy.sparse import csr_matrix


def test_zero_row():
    matrix = bsparse.zero_row(10)

    assert matrix.shape == (1, 10)
    assert matrix.nnz == 0
    assert isinstance(matrix.data, np.ndarray)
    assert isinstance(matrix.indices, np.ndarray)
    assert isinstance(matrix.indptr, np.ndarray)


def test_empty_row():
    matrix = bsparse.empty_row(10)

    assert matrix.shape == (0, 10)
    assert isinstance(matrix.data, np.ndarray)
    assert isinstance(matrix.indices, np.ndarray)
    assert isinstance(matrix.indptr, np.ndarray)


def test_from_array():
    array = np.eye(2)
    spmatrix = bsparse.from_array(array)

    assert spmatrix.shape == (2, 2)
    assert np.all(spmatrix.data == np.array([1, 1]))
    assert np.all(spmatrix.indices == np.array([0, 1]))
    assert np.all(spmatrix.indptr == np.array([0, 1, 2]))


def test_to_array():
    # If we give it an array, it should give back the same array
    array = np.eye(2)
    assert np.all(bsparse.to_array(array) == array)

    # If we give it a spmatrix, it should give the corresponding array
    spmatrix = csr_matrix(array)
    assert np.all(bsparse.to_array(spmatrix) == array)
    assert np.all(bsparse.to_array(spmatrix) == spmatrix.toarray())


def test_is_empty():
    matrix = bsparse.empty_row(10)

    assert bsparse.is_empty(matrix)


def test_is_sparse():
    array = np.array([1, 2, 3])
    matrix = csr_matrix(array)

    assert bsparse.is_sparse(matrix)
    assert not bsparse.is_sparse(array)


def test_is_one():
    matrix = csr_matrix(np.array([1, 0, 1, 1]))

    assert bsparse.is_one(0, matrix)
    assert not bsparse.is_one(1, matrix)
    assert bsparse.is_one(2, matrix)
    assert bsparse.is_one(3, matrix)


def test_insert_mod2():
    matrix = bsparse.zero_row(3)

    bsparse.insert_mod2(2, matrix)

    assert bsparse.is_one(2, matrix)
    assert not bsparse.is_one(0, matrix)
    assert not bsparse.is_one(1, matrix)

    bsparse.insert_mod2(2, matrix)

    assert len(matrix.data) == 0

    bsparse.insert_mod2(0, matrix)
    bsparse.insert_mod2(1, matrix)
    bsparse.insert_mod2(2, matrix)

    assert len(matrix.data) == 3


def test_equal():
    matrix1 = bsparse.from_array([[0, 1, 0, 1, 0, 0], [1, 1, 1, 1, 0, 1]])
    matrix1_prime = bsparse.from_array([[0, 1, 0, 1, 0, 0],
                                        [1, 1, 1, 1, 0, 1]])
    matrix1_wrong = bsparse.from_array([[1, 1, 0, 1, 0, 0],
                                        [1, 1, 1, 1, 0, 1]])
    matrix2 = bsparse.from_array(np.ones((5, 10)))
    matrix3 = bsparse.from_array(np.zeros((5, 10)))

    assert bsparse.equal(matrix1, matrix1_prime)
    assert not bsparse.equal(matrix1, matrix1_wrong)
    assert not bsparse.equal(matrix1, matrix2)
    assert not bsparse.equal(matrix1, matrix3)
    assert not bsparse.equal(matrix1, 1)
    assert not bsparse.equal(1, matrix1)
    assert not bsparse.equal(matrix1, 0)
    assert not bsparse.equal(0, matrix1)

    assert bsparse.equal(matrix2, 1)
    assert bsparse.equal(1, matrix2)
    assert not bsparse.equal(matrix2, 0)
    assert not bsparse.equal(0, matrix2)
    assert bsparse.equal(matrix3, 0)
    assert bsparse.equal(0, matrix3)
    assert not bsparse.equal(matrix3, 1)
    assert not bsparse.equal(1, matrix3)


def test_vstack():
    empty_matrix = bsparse.empty_row(2)
    first_row = bsparse.zero_row(2)
    second_row = bsparse.zero_row(2)

    bsparse.insert_mod2(0, first_row)
    bsparse.insert_mod2(1, second_row)

    still_first_row = bsparse.vstack([empty_matrix, first_row])
    new_matrix = bsparse.vstack([first_row, second_row])

    assert bsparse.equal(still_first_row, first_row)
    assert new_matrix.shape == (2, 2)
    assert np.all(bsparse.to_array(new_matrix) == np.eye(2))


def test_hstack():
    full_block = bsparse.from_array([[0, 1, 0, 1, 0, 0], [1, 1, 1, 1, 0, 1]])
    a_block = bsparse.from_array([[0, 1, 0], [1, 1, 1]])
    b_block = bsparse.from_array([[1, 0, 0], [1, 0, 1]])

    new_matrix = bsparse.hstack([a_block, b_block])

    assert new_matrix.shape == (a_block.shape[0],
                                a_block.shape[1] + b_block.shape[1])
    assert bsparse.equal(new_matrix, full_block)


def test_hsplit():
    # Test with row vectors
    full_row = bsparse.from_array([1, 0, 0, 1])
    a_row = bsparse.from_array([1, 0])
    b_row = bsparse.from_array([0, 1])

    a, b = bsparse.hsplit(full_row)

    assert bsparse.equal(a_row, a)
    assert bsparse.equal(b_row, b)

    # Test with full matrix
    full_block = bsparse.from_array([[0, 1, 0, 1, 0, 0], [1, 1, 1, 1, 0, 1]])
    a_block = bsparse.from_array([[0, 1, 0], [1, 1, 1]])
    b_block = bsparse.from_array([[1, 0, 0], [1, 0, 1]])

    a, b = bsparse.hsplit(full_block)
    a_bis, b_bis = bsparse.hsplit(bsparse.hstack([a_block, b_block]))

    assert bsparse.equal(a_block, a)
    assert bsparse.equal(b_block, b)
    assert bsparse.equal(a_block, a_bis)
    assert bsparse.equal(b_block, b_bis)


def test_dot():
    a = bsparse.zero_row(10)
    b = bsparse.zero_row(10)

    assert bsparse.dot(a, b) == 0
    bsparse.insert_mod2(5, a)
    assert bsparse.dot(a, b) == 0
    bsparse.insert_mod2(5, b)
    assert bsparse.dot(a, b) == 1
    bsparse.insert_mod2(2, a)
    assert bsparse.dot(a, b) == 1
    bsparse.insert_mod2(2, b)
    assert bsparse.dot(a, b) == 0

    # Test that it works with numpy arrays too
    assert bsparse.dot(a.toarray(), b) == 0
    assert bsparse.dot(a.toarray(), b.toarray()) == 0
    assert bsparse.dot(a, b.toarray()) == 0



"""
API for running simulations.
"""

import datetime
import numpy as np
from panqec.codes import StabilizerCode
from panqec.decoders import BaseDecoder
from panqec.error_models import BaseErrorModel
from ..bpauli import get_effective_error
from . import BaseSimulation


def run_once(
    code: StabilizerCode,
    error_model: BaseErrorModel,
    decoder: BaseDecoder,
    error_rate: float,
    rng=None
) -> dict:
    """Run a simulation once and return the results as a dictionary."""

    if not (0 <= error_rate <= 1):
        raise ValueError('Error rate must be in [0, 1].')

    if rng is None:
        rng = np.random.default_rng()

    error = error_model.generate(code, error_rate=error_rate, rng=rng)
    syndrome = code.measure_syndrome(error)
    correction = decoder.decode(syndrome)
    total_error = (correction + error) % 2
    effective_error = get_effective_error(
        total_error, code.logicals_x, code.logicals_z
    )
    codespace = code.in_codespace(total_error)
    success = bool(np.all(effective_error == 0)) and codespace

    results = {
        'error': error,
        'syndrome': syndrome,
        'correction': correction,
        'effective_error': effective_error,
        'success': success,
        'codespace': codespace,
    }

    return results


def calculate_logical_error_rate(
    code: StabilizerCode,
    error_model: BaseErrorModel,
    decoder: BaseDecoder,
    error_rate: float,
    n_runs: int,
    verbose: bool = False
):
    """Simple function to calculate the logical error rate"""

    n_fails = 0
    for run in range(n_runs):
        if verbose:
            print(f"Run {run+1} / {n_runs}", end='\r')
        results = run_once(code, error_model, decoder, error_rate)
        n_fails += 1 - results['success']

    return n_fails / n_runs


class DirectSimulation(BaseSimulation):
    """Quantum Error Correction Simulation.

    Parameters
    -----------
    code : StabilizerCode
        The code to simulate.
    error_model : BaseErrorModel
        The error model to use.
    decoder: BaseDecoder
        The decoder to use.
    error_rate : float
        The error rate parameter.
    compress : bool
        Set False to not compress the output files and save as plain json.
    verbose : bool
        Set False to suppress output.
    rng :
        Set Random number generator if you want to seed it.
    """

    start_time: datetime.datetime
    code: StabilizerCode
    error_model: BaseErrorModel
    decoder: BaseDecoder
    error_rate: float
    label: str
    _results: dict = {}
    rng = None

    def __init__(
        self,
        code: StabilizerCode,
        error_model: BaseErrorModel,
        decoder: BaseDecoder,
        error_rate: float,
        compress: bool = True,
        verbose=True,
        rng=None
    ):
        super().__init__(
            code, error_model, compress=compress, verbose=verbose, rng=rng
        )

        self.decoder = decoder
        self.error_rate = error_rate

        self._results = {
            **self._results,
            'effective_error': [],
            'success': [],
            'codespace': [],
        }
        self._inputs = {
            **self._inputs,
            'decoder': {
                'name': self.decoder.id,
                'parameters': self.decoder.params
            },
            'error_rate': self.error_rate,
            'method': {
                'name': 'direct',
                'parameters': {}
            }
        }

    def _run(self, n_runs: int):
        """Run assuming perfect measurement."""

        for i_run in range(n_runs):
            shot = run_once(
                self.code, self.error_model, self.decoder,
                error_rate=self.error_rate,
                rng=self.rng
            )
            for key, value in shot.items():
                if key in self._results.keys():
                    self._results[key].append(value)

            self._results['n_runs'] += 1

    def get_results(self):
        """Return results as dictionary."""

        success = np.array(self.results['success'])
        if len(success) > 0:
            n_fail = np.sum(~success)
        else:
            n_fail = 0
        simulation_data = {
            'n_success': np.sum(success),
            'n_fail': n_fail,
            'n_runs': len(success),
        }

        # Use sample mean as estimator for effective error rate.
        if simulation_data['n_runs'] != 0:
            simulation_data['p_est'] = (
                simulation_data['n_fail']/simulation_data['n_runs']
            )
        else:
            simulation_data['p_est'] = np.nan

        # Use posterior Beta distribution of the effective error rate
        # standard distribution as standard error.
        simulation_data['p_se'] = np.sqrt(
            simulation_data['p_est']*(1 - simulation_data['p_est'])
            / (simulation_data['n_runs'] + 1)
        )
        return simulation_data



from ._color_666_planar_code import Color666PlanarCode  # noqa
from ._color_666_toric_code import Color666ToricCode  # noqa
from ._color_488_code import Color488Code  # noqa

__all__ = [
    'Color666PlanarCode',
    'Color666ToricCode',
    'Color488Code',
]



from .base._base_decoder import BaseDecoder  # noqa

from .belief_propagation.bposd_decoder import BeliefPropagationOSDDecoder  # noqa
from .belief_propagation.mbp_decoder import MemoryBeliefPropagationDecoder  # noqa

from .matching._matching_decoder import MatchingDecoder  # noqa
from .xcube._xcube_matching_decoder import XCubeMatchingDecoder  # noqa
from .sweepmatch._sweep_decoder_3d import SweepDecoder3D  # noqa
from .sweepmatch._sweep_match_decoder import SweepMatchDecoder  # noqa
from .sweepmatch._rotated_sweep_decoder import RotatedSweepDecoder3D  # noqa
from .sweepmatch._rotated_sweep_match_decoder import RotatedSweepMatchDecoder  # noqa

__all__ = [
    "BaseDecoder",
    "BeliefPropagationOSDDecoder",
    "MemoryBeliefPropagationDecoder",

    "RotatedSweepDecoder3D",
    "RotatedSweepMatchDecoder",
    "SweepDecoder3D",
    "SweepMatchDecoder",
    "MatchingDecoder",
    "XCubeMatchingDecoder"
]



"""
API for running simulations.
"""
import numpy as np
import datetime
from typing import List, Tuple
from panqec.codes import StabilizerCode
from panqec.decoders import BaseDecoder
from panqec.error_models import BaseErrorModel
from . import calculate_logical_error_rate, BaseSimulation


def g(x):
    return 1 / (1 + x)


g = np.vectorize(g)


class SplittingSimulation(BaseSimulation):
    """Quantum Error Correction Simulation."""

    start_time: datetime.datetime
    code: StabilizerCode
    error_model: BaseErrorModel
    decoder: BaseDecoder
    error_rates: np.ndarray
    current_error: List[np.ndarray]  # error vector for each error rate
    label: str
    _results: dict = {}
    rng = None

    def __init__(
        self,
        code: StabilizerCode,
        error_model: BaseErrorModel,
        decoders: List[BaseDecoder],
        error_rates: List[float],
        n_init_runs: int,
        start_run: int = 0,
        compress: bool = True,
        verbose: bool = True,
        rng=None
    ):
        super().__init__(
            code, error_model, compress=compress, verbose=verbose, rng=rng
        )

        self.decoders = decoders
        self.error_rates = np.sort(error_rates)[::-1]
        self.n_init_runs = n_init_runs

        self.current_error = []
        self.initial_logical_p = None
        self.start_run = start_run

        self._results = {
            **self._results,
            'error_rates': self.error_rates,
            'log_p_errors': [[] for _ in self.error_rates],
            'logical_error_rates': []
        }
        self._inputs = {
            **self._inputs,
            'decoder': {
                'name': self.decoders[0].id,
                'parameters': self.decoders[0].params,
            },
            'error_rates': self.error_rates,
            'method': {
                'name': 'splitting',
                'parameters': {
                    'n_init_runs': n_init_runs,
                    'start_run': start_run
                }
            }
        }

    def _run(self, n_runs: int):
        """Run assuming perfect measurement."""
        # Find an error that leads to a decoding failure

        if len(self.current_error) == 0:
            # # Only works for deformed 3D rotated toric code
            # # TODO; remove that (or replace it)
            # initial_error = np.concatenate([np.zeros(self.code.n),
            #                                 np.ones(self.code.n)])
            # deformation_indices = self.error_model.get_deformation_indices(
            #     self.code
            # )
            # initial_error[self.code.n:][deformation_indices] = 0
            # initial_error[:self.code.n][deformation_indices] = 1
            if (self.error_model.error_probability(
                self.code.logicals_x[0], self.code, 0.5
            ) != 0):
                initial_error: np.ndarray = self.code.logicals_x[0]
            elif (self.error_model.error_probability(
                self.code.logicals_z[0], self.code, 0.5
            ) != 0):
                initial_error = self.code.logicals_z[0]
            else:
                raise NotImplementedError(
                    "Splitting method: neither of the logicals has nonzero"
                    "probability. Please specify another error that fails"
                    "when decoded"
                )

            # Check that the chosen error indeed fails
            syndrome = self.code.measure_syndrome(initial_error)
            correction = self.decoders[0].decode(syndrome)
            total_error = (correction + initial_error) % 2
            if self.code.is_success(total_error):
                raise ValueError(
                    "Splitting method: the chosen initial error"
                    "does not fail when decoded"
                )
            self.current_error = [initial_error
                                  for _ in range(len(self.error_rates))]

        for i_run in range(n_runs):
            for i_p, error_rate in enumerate(self.error_rates):
                self.current_error[i_p], log_p_error = self.get_next_error(
                    self.decoders[i_p], error_rate, self.current_error[i_p]
                )
                self._results['log_p_errors'][i_p].append(log_p_error)
            self._results['n_runs'] += 1

    def postprocess(self):
        super().postprocess()

        logical_p = self.compute_logical_probabilities()
        self._results['logical_error_rates'] = logical_p

    def get_results(self):
        """Return results as dictionary."""

        # self.postprocess()

        simulation_data = {
            'size': self.code.size,
            'code': self.code.label,
            'n': self.code.n,
            'k': self.code.k,
            'd': self.code.d,
            'error_model': self.error_model.label,
            'error_rates': self.error_rates,
            'n_runs': self._results['n_runs'],
            'p_est': self._results['logical_error_rates']
        }

        # Use posterior Beta distribution of the effective error rate
        # standard distribution as standard error.
        # TODO: change that to the actual splitting method std
        simulation_data['p_se'] = np.sqrt(
            simulation_data['p_est']*(1 - simulation_data['p_est'])
            / (simulation_data['n_runs'] + 1)
        )
        return simulation_data

    def get_next_error(
        self,
        decoder: BaseDecoder,
        error_rate: float,
        previous_error: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        if not (0 <= error_rate <= 1):
            raise ValueError('Error rate must be in [0, 1].')

        new_edge = np.zeros(2*self.code.n, dtype='uint')
        e_index = np.random.choice(self.code.n)

        pi, px, py, pz = self.error_model.probability_distribution(self.code,
                                                                   error_rate)

        paulis = []
        if px[e_index] != 0:
            paulis.append('X')
        if py[e_index] != 0:
            paulis.append('Y')
        if pz[e_index] != 0:
            paulis.append('Z')

        e_pauli = np.random.choice(paulis)

        if e_pauli == 'X' or e_pauli == 'Y':
            new_edge[e_index] = 1
        if e_pauli == 'Z' or e_pauli == 'Y':
            new_edge[self.code.n + e_index] = 1

        new_error = (previous_error + new_edge) % 2

        log_p_previous_error = self.error_model.error_probability(
            previous_error, self.code, error_rate, log_output=True
        )
        log_p_new_error = self.error_model.error_probability(
            new_error, self.code, error_rate, log_output=True
        )

        q = np.exp(min(0, log_p_new_error - log_p_previous_error))
        b = np.random.choice([0, 1], p=[1-q, q])

        next_error = previous_error
        log_p_next_error = log_p_previous_error

        if b:
            syndrome = self.code.measure_syndrome(new_error)
            correction = decoder.decode(syndrome)
            total_error = (correction + new_error) % 2
            if (self.code.is_logical_error(total_error)
                    or not self.code.in_codespace(total_error)):
                next_error = new_error
                log_p_next_error = log_p_new_error

        return next_error, log_p_next_error

    def compute_optimal_c(self):
        list_c = np.linspace(0.0001, 1, 100)
        n_p = len(self.error_rates)
        log_p_errors = np.array(self._results['log_p_errors'])

        optimal_c = []

        for j in range(n_p-1):
            print(f"{j+1} / {n_p - 1}")
            lhs = []
            rhs = []
            for c in list_c:
                print(f"C : {c}", end="\r")
                lhs.append(np.sum(
                    g(c * np.exp(log_p_errors[j][self.start_run:]
                      - log_p_errors[j+1][self.start_run:]))
                ))
                rhs.append(np.sum(
                    g(1/c * np.exp(log_p_errors[j+1][self.start_run:]
                      - log_p_errors[j][self.start_run:]))
                ))

            lhs = np.array(lhs)
            rhs = np.array(rhs)

            try:
                idx = np.argwhere(np.diff(np.sign(lhs - rhs))).flatten()[0]
                optimal_c.append(list_c[idx])
            except IndexError:
                optimal_c.append(1)

        return optimal_c

    def compute_logical_probabilities(self):
        n_p = len(self.error_rates)
        log_p_errors = np.array(self._results['log_p_errors'])

        logical_p = np.zeros(n_p)

        if self.verbose:
            print("Compute initial logical error rate")

        self.initial_logical_p = calculate_logical_error_rate(
            self.code, self.error_model, self.decoders[0],
            self.error_rates[0], self.n_init_runs, verbose=True
        )
        logical_p[0] = self.initial_logical_p

        if self.verbose:
            print(f"P = {self.initial_logical_p}")

        if self.verbose:
            print("Compute optimal C")
        optimal_c = self.compute_optimal_c()

        if self.verbose:
            print("Compute logical error rates")

        for j in range(n_p-1):
            c = optimal_c[j]
            numerator = np.sum(
                g(c * np.exp(log_p_errors[j][self.start_run:]
                             - log_p_errors[j+1][self.start_run:]))
            )
            denominator = np.sum(
                g(1/c * np.exp(log_p_errors[j+1][self.start_run:]
                               - log_p_errors[j][self.start_run:]))
            )
            ratio = c * numerator / denominator

            logical_p[j+1] = logical_p[j] * ratio

        return logical_p
